#!/usr/bin/python
#
# @author: Gaurav Rastogi (grastogi@avinetworks.com)
#          Eric Anderson (eanderson@avinetworks.com)
# module_check: supported
# Avi Version: 17.1.1
#
# Copyright: (c) 2017 Gaurav Rastogi, <grastogi@avinetworks.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
#

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = '''
---
module: avi_sslprofile
author: Gaurav Rastogi (grastogi@avinetworks.com)

short_description: Module for setup of SSLProfile Avi RESTful Object
description:
    - This module is used to configure SSLProfile object
    - more examples at U(https://github.com/avinetworks/devops)
requirements: [ avisdk ]
version_added: "2.3"
options:
    state:
        description:
            - The state that should be applied on the entity.
        default: present
        choices: ["absent", "present"]
    avi_api_update_method:
        description:
            - Default method for object update is HTTP PUT.
            - Setting to patch will override that behavior to use HTTP PATCH.
        version_added: "2.5"
        default: put
        choices: ["put", "patch"]
    avi_api_patch_op:
        description:
            - Patch operation to use when using avi_api_update_method as patch.
        version_added: "2.5"
        choices: ["add", "replace", "delete"]
    accepted_ciphers:
        description:
            - Ciphers suites represented as defined by U(http://www.openssl.org/docs/apps/ciphers.html).
            - Default value when not specified in API or module is interpreted by Avi Controller as AES:3DES:RC4.
    accepted_versions:
        description:
            - Set of versions accepted by the server.
    cipher_enums:
        description:
            - Enum options - tls_ecdhe_ecdsa_with_aes_128_gcm_sha256, tls_ecdhe_ecdsa_with_aes_256_gcm_sha384, tls_ecdhe_rsa_with_aes_128_gcm_sha256,
            - tls_ecdhe_rsa_with_aes_256_gcm_sha384, tls_ecdhe_ecdsa_with_aes_128_cbc_sha256, tls_ecdhe_ecdsa_with_aes_256_cbc_sha384,
            - tls_ecdhe_rsa_with_aes_128_cbc_sha256, tls_ecdhe_rsa_with_aes_256_cbc_sha384, tls_rsa_with_aes_128_gcm_sha256, tls_rsa_with_aes_256_gcm_sha384,
            - tls_rsa_with_aes_128_cbc_sha256, tls_rsa_with_aes_256_cbc_sha256, tls_ecdhe_ecdsa_with_aes_128_cbc_sha, tls_ecdhe_ecdsa_with_aes_256_cbc_sha,
            - tls_ecdhe_rsa_with_aes_128_cbc_sha, tls_ecdhe_rsa_with_aes_256_cbc_sha, tls_rsa_with_aes_128_cbc_sha, tls_rsa_with_aes_256_cbc_sha,
            - tls_rsa_with_3des_ede_cbc_sha, tls_rsa_with_rc4_128_sha.
    description:
        description:
            - User defined description for the object.
    dhparam:
        description:
            - Dh parameters used in ssl.
            - At this time, it is not configurable and is set to 2048 bits.
    enable_ssl_session_reuse:
        description:
            - Enable ssl session re-use.
            - Default value when not specified in API or module is interpreted by Avi Controller as True.
        type: bool
    name:
        description:
            - Name of the object.
        required: true
    prefer_client_cipher_ordering:
        description:
            - Prefer the ssl cipher ordering presented by the client during the ssl handshake over the one specified in the ssl profile.
            - Default value when not specified in API or module is interpreted by Avi Controller as False.
        type: bool
    send_close_notify:
        description:
            - Send 'close notify' alert message for a clean shutdown of the ssl connection.
            - Default value when not specified in API or module is interpreted by Avi Controller as True.
        type: bool
    ssl_rating:
        description:
            - Sslrating settings for sslprofile.
    ssl_session_timeout:
        description:
            - The amount of time before an ssl session expires.
            - Default value when not specified in API or module is interpreted by Avi Controller as 86400.
            - Units(SEC).
    tags:
        description:
            - List of tag.
    tenant_ref:
        description:
            - It is a reference to an object of type tenant.
    type:
        description:
            - Ssl profile type.
            - Enum options - SSL_PROFILE_TYPE_APPLICATION, SSL_PROFILE_TYPE_SYSTEM.
            - Field introduced in 17.2.8.
            - Default value when not specified in API or module is interpreted by Avi Controller as SSL_PROFILE_TYPE_APPLICATION.
        version_added: "2.6"
    url:
        description:
            - Avi controller URL of the object.
    uuid:
        description:
            - Unique object identifier of the object.
extends_documentation_fragment:
    - avi
'''

EXAMPLES = """
  - name: Create SSL profile with list of allowed ciphers
    avi_sslprofile:
      controller: '{{ controller }}'
      username: '{{ username }}'
      password: '{{ password }}'
      accepted_ciphers: >
        ECDHE-ECDSA-AES128-GCM-SHA256:ECDHE-ECDSA-AES128-SHA:ECDHE-ECDSA-AES256-SHA:
        ECDHE-ECDSA-AES256-GCM-SHA384:ECDHE-ECDSA-AES128-SHA256:ECDHE-ECDSA-AES256-SHA384:
        AES128-GCM-SHA256:AES256-GCM-SHA384:AES128-SHA256:AES256-SHA256:AES128-SHA:
        AES256-SHA:DES-CBC3-SHA:ECDHE-RSA-AES128-SHA:ECDHE-RSA-AES256-SHA384:
        ECDHE-RSA-AES128-SHA256:ECDHE-RSA-AES256-GCM-SHA384:ECDHE-RSA-AES128-GCM-SHA256:ECDHE-RSA-AES256-SHA
      accepted_versions:
      - type: SSL_VERSION_TLS1
      - type: SSL_VERSION_TLS1_1
      - type: SSL_VERSION_TLS1_2
      cipher_enums:
      - TLS_ECDHE_ECDSA_WITH_AES_128_GCM_SHA256
      - TLS_ECDHE_ECDSA_WITH_AES_128_CBC_SHA
      - TLS_ECDHE_ECDSA_WITH_AES_256_CBC_SHA
      - TLS_ECDHE_ECDSA_WITH_AES_256_GCM_SHA384
      - TLS_ECDHE_ECDSA_WITH_AES_128_CBC_SHA256
      - TLS_ECDHE_ECDSA_WITH_AES_256_CBC_SHA384
      - TLS_RSA_WITH_AES_128_GCM_SHA256
      - TLS_RSA_WITH_AES_256_GCM_SHA384
      - TLS_RSA_WITH_AES_128_CBC_SHA256
      - TLS_RSA_WITH_AES_256_CBC_SHA256
      - TLS_RSA_WITH_AES_128_CBC_SHA
      - TLS_RSA_WITH_AES_256_CBC_SHA
      - TLS_RSA_WITH_3DES_EDE_CBC_SHA
      - TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA
      - TLS_ECDHE_RSA_WITH_AES_256_CBC_SHA384
      - TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA256
      - TLS_ECDHE_RSA_WITH_AES_256_GCM_SHA384
      - TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256
      - TLS_ECDHE_RSA_WITH_AES_256_CBC_SHA
      name: PFS-BOTH-RSA-EC
      send_close_notify: true
      ssl_rating:
        compatibility_rating: SSL_SCORE_EXCELLENT
        performance_rating: SSL_SCORE_EXCELLENT
        security_score: '100.0'
      tenant_ref: Demo
"""

RETURN = '''
obj:
    description: SSLProfile (api/sslprofile) object
    returned: success, changed
    type: dict
'''

from ansible.module_utils.basic import AnsibleModule
try:
    from ansible.module_utils.network.avi.avi import (
        avi_common_argument_spec, HAS_AVI, avi_ansible_api)
except ImportError:
    HAS_AVI = False


def main():
    argument_specs = dict(
        state=dict(default='present',
                   choices=['absent', 'present']),
        avi_api_update_method=dict(default='put',
                                   choices=['put', 'patch']),
        avi_api_patch_op=dict(choices=['add', 'replace', 'delete']),
        accepted_ciphers=dict(type='str',),
        accepted_versions=dict(type='list',),
        cipher_enums=dict(type='list',),
        description=dict(type='str',),
        dhparam=dict(type='str',),
        enable_ssl_session_reuse=dict(type='bool',),
        name=dict(type='str', required=True),
        prefer_client_cipher_ordering=dict(type='bool',),
        send_close_notify=dict(type='bool',),
        ssl_rating=dict(type='dict',),
        ssl_session_timeout=dict(type='int',),
        tags=dict(type='list',),
        tenant_ref=dict(type='str',),
        type=dict(type='str',),
        url=dict(type='str',),
        uuid=dict(type='str',),
    )
    argument_specs.update(avi_common_argument_spec())
    module = AnsibleModule(
        argument_spec=argument_specs, supports_check_mode=True)
    if not HAS_AVI:
        return module.fail_json(msg=(
            'Avi python API SDK (avisdk>=17.1) is not installed. '
            'For more details visit https://github.com/avinetworks/sdk.'))
    return avi_ansible_api(module, 'sslprofile',
                           set([]))

if __name__ == '__main__':
    main()



# (c) 2015 Toshio Kuratomi <tkuratomi@ansible.com>
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

import jinja2
from ansible.compat.tests import unittest

from ansible.template import _escape_backslashes, _count_newlines_from_end

# These are internal utility functions only needed for templating.  They're
# algorithmic so good candidates for unittesting by themselves


class TestBackslashEscape(unittest.TestCase):

    test_data = (
        # Test backslashes in a filter arg are double escaped
        dict(
            template=u"{{ 'test2 %s' | format('\\1') }}",
            intermediate=u"{{ 'test2 %s' | format('\\\\1') }}",
            expectation=u"test2 \\1",
            args=dict()
        ),
        # Test backslashes inside the jinja2 var itself are double
        # escaped
        dict(
            template=u"Test 2\\3: {{ '\\1 %s' | format('\\2') }}",
            intermediate=u"Test 2\\3: {{ '\\\\1 %s' | format('\\\\2') }}",
            expectation=u"Test 2\\3: \\1 \\2",
            args=dict()
        ),
        # Test backslashes outside of the jinja2 var are not double
        # escaped
        dict(
            template=u"Test 2\\3: {{ 'test2 %s' | format('\\1') }}; \\done",
            intermediate=u"Test 2\\3: {{ 'test2 %s' | format('\\\\1') }}; \\done",
            expectation=u"Test 2\\3: test2 \\1; \\done",
            args=dict()
        ),
        # Test backslashes in a variable sent to a filter are handled
        dict(
            template=u"{{ 'test2 %s' | format(var1) }}",
            intermediate=u"{{ 'test2 %s' | format(var1) }}",
            expectation=u"test2 \\1",
            args=dict(var1=u'\\1')
        ),
        # Test backslashes in a variable expanded by jinja2 are double
        # escaped
        dict(
            template=u"Test 2\\3: {{ var1 | format('\\2') }}",
            intermediate=u"Test 2\\3: {{ var1 | format('\\\\2') }}",
            expectation=u"Test 2\\3: \\1 \\2",
            args=dict(var1=u'\\1 %s')
        ),
    )

    def setUp(self):
        self.env = jinja2.Environment()

    def tearDown(self):
        pass

    def test_backslash_escaping(self):

        for test in self.test_data:
            intermediate = _escape_backslashes(test['template'], self.env)
            self.assertEquals(intermediate, test['intermediate'])
            template = jinja2.Template(intermediate)
            args = test['args']
            self.assertEquals(template.render(**args), test['expectation'])


class TestCountNewlines(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_zero_length_string(self):
        self.assertEquals(_count_newlines_from_end(u''), 0)

    def test_short_string(self):
        self.assertEquals(_count_newlines_from_end(u'The quick\n'), 1)

    def test_one_newline(self):
        self.assertEquals(_count_newlines_from_end(u'The quick brown fox jumped over the lazy dog' * 1000 + u'\n'), 1)

    def test_multiple_newlines(self):
        self.assertEquals(_count_newlines_from_end(u'The quick brown fox jumped over the lazy dog' * 1000 + u'\n\n\n'), 3)

    def test_zero_newlines(self):
        self.assertEquals(_count_newlines_from_end(u'The quick brown fox jumped over the lazy dog' * 1000), 0)

    def test_all_newlines(self):
        self.assertEquals(_count_newlines_from_end(u'\n' * 10), 10)

    def test_mostly_newlines(self):
        self.assertEquals(_count_newlines_from_end(u'The quick brown fox jumped over the lazy dog' + u'\n' * 1000), 1000)



# -*- coding: utf-8 -*-

# Copyright 2017 Sloane Hertel <shertel@redhat.com>
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

import pytest
import datetime

# Just to test that we have the prerequisite for InventoryModule and instance_data_filter_to_boto_attr
boto3 = pytest.importorskip('boto3')
botocore = pytest.importorskip('botocore')

from ansible.errors import AnsibleError, AnsibleParserError
from ansible.plugins.inventory.aws_ec2 import InventoryModule
from ansible.plugins.inventory.aws_ec2 import instance_data_filter_to_boto_attr

instances = {
    u'Instances': [
        {u'Monitoring': {u'State': 'disabled'},
         u'PublicDnsName': 'ec2-12-345-67-890.compute-1.amazonaws.com',
         u'State': {u'Code': 16, u'Name': 'running'},
         u'EbsOptimized': False,
         u'LaunchTime': datetime.datetime(2017, 10, 31, 12, 59, 25),
         u'PublicIpAddress': '12.345.67.890',
         u'PrivateIpAddress': '098.76.54.321',
         u'ProductCodes': [],
         u'VpcId': 'vpc-12345678',
         u'StateTransitionReason': '',
         u'InstanceId': 'i-00000000000000000',
         u'EnaSupport': True,
         u'ImageId': 'ami-12345678',
         u'PrivateDnsName': 'ip-098-76-54-321.ec2.internal',
         u'KeyName': 'testkey',
         u'SecurityGroups': [{u'GroupName': 'default', u'GroupId': 'sg-12345678'}],
         u'ClientToken': '',
         u'SubnetId': 'subnet-12345678',
         u'InstanceType': 't2.micro',
         u'NetworkInterfaces': [
            {u'Status': 'in-use',
             u'MacAddress': '12:a0:50:42:3d:a4',
             u'SourceDestCheck': True,
             u'VpcId': 'vpc-12345678',
             u'Description': '',
             u'NetworkInterfaceId': 'eni-12345678',
             u'PrivateIpAddresses': [
                 {u'PrivateDnsName': 'ip-098-76-54-321.ec2.internal',
                  u'PrivateIpAddress': '098.76.54.321',
                  u'Primary': True,
                  u'Association':
                      {u'PublicIp': '12.345.67.890',
                       u'PublicDnsName': 'ec2-12-345-67-890.compute-1.amazonaws.com',
                       u'IpOwnerId': 'amazon'}}],
             u'PrivateDnsName': 'ip-098-76-54-321.ec2.internal',
             u'Attachment':
                 {u'Status': 'attached',
                  u'DeviceIndex': 0,
                  u'DeleteOnTermination': True,
                  u'AttachmentId': 'eni-attach-12345678',
                  u'AttachTime': datetime.datetime(2017, 10, 31, 12, 59, 25)},
             u'Groups': [
                 {u'GroupName': 'default',
                  u'GroupId': 'sg-12345678'}],
             u'Ipv6Addresses': [],
             u'OwnerId': '123456789000',
             u'PrivateIpAddress': '098.76.54.321',
             u'SubnetId': 'subnet-12345678',
             u'Association':
                {u'PublicIp': '12.345.67.890',
                 u'PublicDnsName': 'ec2-12-345-67-890.compute-1.amazonaws.com',
                 u'IpOwnerId': 'amazon'}}],
         u'SourceDestCheck': True,
         u'Placement':
            {u'Tenancy': 'default',
             u'GroupName': '',
             u'AvailabilityZone': 'us-east-1c'},
         u'Hypervisor': 'xen',
         u'BlockDeviceMappings': [
            {u'DeviceName': '/dev/xvda',
             u'Ebs':
                {u'Status': 'attached',
                 u'DeleteOnTermination': True,
                 u'VolumeId': 'vol-01234567890000000',
                 u'AttachTime': datetime.datetime(2017, 10, 31, 12, 59, 26)}}],
         u'Architecture': 'x86_64',
         u'RootDeviceType': 'ebs',
         u'RootDeviceName': '/dev/xvda',
         u'VirtualizationType': 'hvm',
         u'Tags': [{u'Value': 'test', u'Key': 'ansible'}, {u'Value': 'aws_ec2', u'Key': 'name'}],
         u'AmiLaunchIndex': 0}],
    u'ReservationId': 'r-01234567890000000',
    u'Groups': [],
    u'OwnerId': '123456789000'
}


@pytest.fixture(scope="module")
def inventory():
    return InventoryModule()


def test_compile_values(inventory):
    found_value = instances['Instances'][0]
    chain_of_keys = instance_data_filter_to_boto_attr['instance.group-id']
    for attr in chain_of_keys:
        found_value = inventory._compile_values(found_value, attr)
    assert found_value == "sg-12345678"


def test_get_boto_attr_chain(inventory):
    instance = instances['Instances'][0]
    assert inventory._get_boto_attr_chain('network-interface.addresses.private-ip-address', instance) == "098.76.54.321"


def test_boto3_conn(inventory):
    inventory._options = {"boto_profile": "first_precedence",
                          "aws_access_key_id": "test_access_key",
                          "aws_secret_access_key": "test_secret_key",
                          "aws_security_token": "test_security_token"}
    inventory._set_credentials()
    with pytest.raises(AnsibleError) as error_message:
        for connection, region in inventory._boto3_conn(regions=['us-east-1']):
            assert error_message == "Insufficient credentials found."


def test_get_hostname_default(inventory):
    instance = instances['Instances'][0]
    assert inventory._get_hostname(instance, hostnames=None) == "ec2-12-345-67-890.compute-1.amazonaws.com"


def test_get_hostname(inventory):
    hostnames = ['ip-address', 'dns-name']
    instance = instances['Instances'][0]
    assert inventory._get_hostname(instance, hostnames) == "12.345.67.890"


def test_set_credentials(inventory):
    inventory._options = {'aws_access_key_id': 'test_access_key',
                          'aws_secret_access_key': 'test_secret_key',
                          'aws_security_token': 'test_security_token',
                          'boto_profile': 'test_profile'}
    inventory._set_credentials()

    assert inventory.boto_profile == "test_profile"
    assert inventory.aws_access_key_id == "test_access_key"
    assert inventory.aws_secret_access_key == "test_secret_key"
    assert inventory.aws_security_token == "test_security_token"


def test_insufficient_credentials(inventory):
    inventory._options = {
        'aws_access_key_id': None,
        'aws_secret_access_key': None,
        'aws_security_token': None,
        'boto_profile': None
    }
    with pytest.raises(AnsibleError) as error_message:
        inventory._set_credentials()
        assert "Insufficient boto credentials found" in error_message


def test_validate_option(inventory):
    assert ['us-east-1'] == inventory._validate_option('regions', list, 'us-east-1')
    assert ['us-east-1'] == inventory._validate_option('regions', list, ['us-east-1'])


def test_illegal_option(inventory):
    bad_filters = [{'tag:Environment': 'dev'}]
    with pytest.raises(AnsibleParserError) as error_message:
        inventory._validate_option('filters', dict, bad_filters)
        assert "The option filters ([{'tag:Environment': 'dev'}]) must be a <class 'dict'>" == error_message


def test_empty_config_query_options(inventory):
    regions, filters, hostnames, strict_permissions = inventory._get_query_options({})
    assert regions == filters == hostnames == []
    assert strict_permissions is True


def test_conig_query_options(inventory):
    regions, filters, hostnames, strict_permissions = inventory._get_query_options(
        {'regions': ['us-east-1', 'us-east-2'],
         'filters': {'tag:Environment': ['dev', 'prod']},
         'hostnames': 'ip-address',
         'strict_permissions': False}
    )
    assert regions == ['us-east-1', 'us-east-2']
    assert filters == [{'Name': 'tag:Environment', 'Values': ['dev', 'prod']}]
    assert hostnames == ['ip-address']
    assert strict_permissions is False


def test_verify_file_bad_config(inventory):
    assert inventory.verify_file('not_aws_config.yml') is False



#
# (c) 2017 Red Hat Inc.
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

from abc import ABCMeta, abstractmethod
from functools import wraps

from ansible.errors import AnsibleError
from ansible.module_utils.six import with_metaclass
from ansible.module_utils._text import to_bytes

try:
    from ncclient.operations import RPCError
    from ncclient.xml_ import to_xml, to_ele
except ImportError:
    raise AnsibleError("ncclient is not installed")

try:
    from lxml.etree import Element, SubElement, tostring, fromstring
except ImportError:
    from xml.etree.ElementTree import Element, SubElement, tostring, fromstring


def ensure_connected(func):
    @wraps(func)
    def wrapped(self, *args, **kwargs):
        if not self._connection._connected:
            self._connection._connect()
        return func(self, *args, **kwargs)
    return wrapped


class NetconfBase(with_metaclass(ABCMeta, object)):
    """
    A base class for implementing Netconf connections

    .. note:: Unlike most of Ansible, nearly all strings in
        :class:`TerminalBase` plugins are byte strings.  This is because of
        how close to the underlying platform these plugins operate.  Remember
        to mark literal strings as byte string (``b"string"``) and to use
        :func:`~ansible.module_utils._text.to_bytes` and
        :func:`~ansible.module_utils._text.to_text` to avoid unexpected
        problems.

        List of supported rpc's:
            :get: Retrieves running configuration and device state information
            :get_config: Retrieves the specified configuration from the device
            :edit_config: Loads the specified commands into the remote device
            :commit: Load configuration from candidate to running
            :discard_changes: Discard changes to candidate datastore
            :validate: Validate the contents of the specified configuration.
            :lock: Allows the client to lock the configuration system of a device.
            :unlock: Release a configuration lock, previously obtained with the lock operation.
            :copy_config: create or replace an entire configuration datastore with the contents of another complete
                          configuration datastore.
            :get-schema: Retrieves the required schema from the device
            :get_capabilities: Retrieves device information and supported rpc methods

            For JUNOS:
            :execute_rpc: RPC to be execute on remote device
            :load_configuration: Loads given configuration on device

        Note: rpc support depends on the capabilites of remote device.

        :returns: Returns output received from remote device as byte string
        Note: the 'result' or 'error' from response should to be converted to object
              of ElementTree using 'fromstring' to parse output as xml doc

              'get_capabilities()' returns 'result' as a json string.

            Usage:
            from ansible.module_utils.connection import Connection

            conn = Connection()
            data = conn.execute_rpc(rpc)
            reply = fromstring(reply)

            data = conn.get_capabilities()
            json.loads(data)

            conn.load_configuration(config=[''set system ntp server 1.1.1.1''], action='set', format='text')
    """

    def __init__(self, connection):
        self._connection = connection
        self.m = self._connection._manager

    @ensure_connected
    def rpc(self, name):
        """RPC to be execute on remote device
           :name: Name of rpc in string format"""
        try:
            obj = to_ele(name)
            resp = self.m.rpc(obj)
            return resp.data_xml if hasattr(resp, 'data_xml') else resp.xml
        except RPCError as exc:
            msg = exc.xml
            raise Exception(to_xml(msg))

    @ensure_connected
    def get_config(self, source=None, filter=None):
        """Retrieve all or part of a specified configuration
           (by default entire configuration is retrieved).

        :param source: Name of the configuration datastore being queried, defaults to running datastore
        :param filter: This argument specifies the portion of the configuration data to retrieve
        :return: Returns xml string containing the RPC response received from remote host
        """
        if isinstance(filter, list):
            filter = tuple(filter)

        if not source:
            source = 'running'
        resp = self.m.get_config(source=source, filter=filter)
        return resp.data_xml if hasattr(resp, 'data_xml') else resp.xml

    @ensure_connected
    def get(self, filter=None):
        """Retrieve device configuration and state information.

        :param filter: This argument specifies the portion of the state data to retrieve
                        (by default entire state data is retrieved)
        :return: Returns xml string containing the RPC response received from remote host
        """
        if isinstance(filter, list):
            filter = tuple(filter)
        resp = self.m.get(filter=filter)
        response = resp.data_xml if hasattr(resp, 'data_xml') else resp.xml
        return response

    @ensure_connected
    def edit_config(self, *args, **kwargs):
        """Loads all or part of the specified *config* to the *target* configuration datastore.

            :target: is the name of the configuration datastore being edited
            :config: is the configuration, which must be rooted in the `config` element.
                        It can be specified either as a string or an :class:`~xml.etree.ElementTree.Element`.
            :default_operation: if specified must be one of { `"merge"`, `"replace"`, or `"none"` }
            :test_option: if specified must be one of { `"test_then_set"`, `"set"` }
            :error_option: if specified must be one of { `"stop-on-error"`, `"continue-on-error"`, `"rollback-on-error"` }
            The `"rollback-on-error"` *error_option* depends on the `:rollback-on-error` capability.
        """
        resp = self.m.edit_config(*args, **kwargs)
        return resp.data_xml if hasattr(resp, 'data_xml') else resp.xml

    @ensure_connected
    def validate(self, *args, **kwargs):
        """Validate the contents of the specified configuration.
        :source: is the name of the configuration datastore being validated or `config`
        element containing the configuration subtree to be validated
        """
        resp = self.m.validate(*args, **kwargs)
        return resp.data_xml if hasattr(resp, 'data_xml') else resp.xml

    @ensure_connected
    def copy_config(self, *args, **kwargs):
        """Create or replace an entire configuration datastore with the contents of another complete
        configuration datastore.
        :source: is the name of the configuration datastore to use as the source of the
                 copy operation or `config` element containing the configuration subtree to copy
        :target: is the name of the configuration datastore to use as the destination of the copy operation"""
        resp = self.m.copy_config(*args, **kwargs)
        return resp.data_xml if hasattr(resp, 'data_xml') else resp.xml

    @ensure_connected
    def dispatch(self, request):
        """Execute operation on the remote device
        :request: is the rpc request including attributes as XML string
        """
        req = fromstring(request)
        resp = self.m.dispatch(req)
        return resp.data_xml if hasattr(resp, 'data_xml') else resp.xml

    @ensure_connected
    def lock(self, target=None):
        """
        Allows the client to lock the configuration system of a device.
        :param target: is the name of the configuration datastore to lock,
                        defaults to candidate datastore
        :return: Returns xml string containing the RPC response received from remote host
        """
        if not target:
            target = 'candidate'
        resp = self.m.lock(target=target)
        return resp.data_xml if hasattr(resp, 'data_xml') else resp.xml

    @ensure_connected
    def unlock(self, target=None):
        """
        Release a configuration lock, previously obtained with the lock operation.
        :param target: is the name of the configuration datastore to unlock,
                       defaults to candidate datastore
        :return: Returns xml string containing the RPC response received from remote host
        """
        """Release a configuration lock, previously obtained with the lock operation.
        :target: is the name of the configuration datastore to unlock
        """
        if not target:
            target = 'candidate'
        resp = self.m.unlock(target=target)
        return resp.data_xml if hasattr(resp, 'data_xml') else resp.xml

    @ensure_connected
    def discard_changes(self):
        """
        Revert the candidate configuration to the currently running configuration.
        Any uncommitted changes are discarded.
        :return: Returns xml string containing the RPC response received from remote host
        """
        resp = self.m.discard_changes()
        return resp.data_xml if hasattr(resp, 'data_xml') else resp.xml

    @ensure_connected
    def commit(self, *args, **kwargs):
        """Commit the candidate configuration as the device's new current configuration.
           Depends on the `:candidate` capability.
           A confirmed commit (i.e. if *confirmed* is `True`) is reverted if there is no
           followup commit within the *timeout* interval. If no timeout is specified the
           confirm timeout defaults to 600 seconds (10 minutes).
           A confirming commit may have the *confirmed* parameter but this is not required.
           Depends on the `:confirmed-commit` capability.
        :confirmed: whether this is a confirmed commit
        :timeout: specifies the confirm timeout in seconds
        """
        resp = self.m.commit(*args, **kwargs)
        return resp.data_xml if hasattr(resp, 'data_xml') else resp.xml

    @ensure_connected
    def get_schema(self, *args, **kwargs):
        """Retrieves the required schema from the device
        """
        resp = self.m.get_schema(*args, **kwargs)
        return resp.data_xml if hasattr(resp, 'data_xml') else resp.xml

    @ensure_connected
    def locked(self, *args, **kwargs):
        resp = self.m.locked(*args, **kwargs)
        return resp.data_xml if hasattr(resp, 'data_xml') else resp.xml

    @abstractmethod
    def get_capabilities(self):
        """Retrieves device information and supported
        rpc methods by device platform and return result
        as a string
        """
        pass

    @staticmethod
    def guess_network_os(obj):
        """Identifies the operating system of
            network device.
        """
        pass

    def get_base_rpc(self):
        """Returns list of base rpc method supported by remote device"""
        return ['get_config', 'edit_config', 'get_capabilities', 'get']

    def put_file(self, source, destination):
        """Copies file over scp to remote device"""
        pass

    def fetch_file(self, source, destination):
        """Fetch file over scp from remote device"""
        pass

    def get_device_operations(self, server_capabilities):
        operations = {}
        capabilities = '\n'.join(server_capabilities)
        operations['supports_commit'] = ':candidate' in capabilities
        operations['supports_defaults'] = ':with-defaults' in capabilities
        operations['supports_confirm_commit'] = ':confirmed-commit' in capabilities
        operations['supports_startup'] = ':startup' in capabilities
        operations['supports_xpath'] = ':xpath' in capabilities
        operations['supports_writable_running'] = ':writable-running' in capabilities

        operations['lock_datastore'] = []
        if operations['supports_writable_running']:
            operations['lock_datastore'].append('running')

        if operations['supports_commit']:
            operations['lock_datastore'].append('candidate')

        if operations['supports_startup']:
            operations['lock_datastore'].append('startup')

        operations['supports_lock'] = True if len(operations['lock_datastore']) else False

        return operations

# TODO Restore .xml, when ncclient supports it for all platforms



#!/usr/bin/python
#
# This is a free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This Ansible library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this library.  If not, see <http://www.gnu.org/licenses/>.

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['stableinterface'],
                    'supported_by': 'core'}


DOCUMENTATION = '''
---
module: ec2_vpc_net_facts
short_description: Gather facts about ec2 VPCs in AWS
description:
    - Gather facts about ec2 VPCs in AWS
version_added: "2.1"
author: "Rob White (@wimnat)"
requirements:
  - boto3
  - botocore
options:
  vpc_ids:
    description:
      - A list of VPC IDs that exist in your account.
    version_added: "2.5"
  filters:
    description:
      - A dict of filters to apply. Each dict item consists of a filter key and a filter value.
        See U(http://docs.aws.amazon.com/AWSEC2/latest/APIReference/API_DescribeVpcs.html) for possible filters.
extends_documentation_fragment:
    - aws
    - ec2
'''

EXAMPLES = '''
# Note: These examples do not set authentication details, see the AWS Guide for details.

# Gather facts about all VPCs
- ec2_vpc_net_facts:

# Gather facts about a particular VPC using VPC ID
- ec2_vpc_net_facts:
    vpc_ids: vpc-00112233

# Gather facts about any VPC with a tag key Name and value Example
- ec2_vpc_net_facts:
    filters:
      "tag:Name": Example

'''

RETURN = '''
vpcs:
    description: Returns an array of complex objects as described below.
    returned: success
    type: complex
    contains:
        id:
            description: The ID of the VPC (for backwards compatibility).
            returned: always
            type: string
        vpc_id:
            description: The ID of the VPC .
            returned: always
            type: string
        state:
            description: The state of the VPC.
            returned: always
            type: string
        tags:
            description: A dict of tags associated with the VPC.
            returned: always
            type: dict
        instance_tenancy:
            description: The instance tenancy setting for the VPC.
            returned: always
            type: string
        is_default:
            description: True if this is the default VPC for account.
            returned: always
            type: boolean
        cidr_block:
            description: The IPv4 CIDR block assigned to the VPC.
            returned: always
            type: string
        classic_link_dns_supported:
            description: True/False depending on attribute setting for classic link DNS support.
            returned: always
            type: boolean
        classic_link_enabled:
            description: True/False depending on if classic link support is enabled.
            returned: always
            type: boolean
        enable_dns_hostnames:
            description: True/False depending on attribute setting for DNS hostnames support.
            returned: always
            type: boolean
        enable_dns_support:
            description: True/False depending on attribute setting for DNS support.
            returned: always
            type: boolean
        ipv6_cidr_block_association_set:
            description: An array of IPv6 cidr block association set information.
            returned: always
            type: complex
            contains:
                association_id:
                    description: The association ID
                    returned: always
                    type: string
                ipv6_cidr_block:
                    description: The IPv6 CIDR block that is associated with the VPC.
                    returned: always
                    type: string
                ipv6_cidr_block_state:
                    description: A hash/dict that contains a single item. The state of the cidr block association.
                    returned: always
                    type: dict
                    contains:
                        state:
                            description: The CIDR block association state.
                            returned: always
                            type: string
'''

import traceback
from ansible.module_utils._text import to_native
from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.ec2 import (
    boto3_conn,
    ec2_argument_spec,
    get_aws_connection_info,
    AWSRetry,
    HAS_BOTO3,
    boto3_tag_list_to_ansible_dict,
    camel_dict_to_snake_dict,
    ansible_dict_to_boto3_filter_list
)

try:
    import botocore
except ImportError:
    pass  # caught by imported HAS_BOTO3


@AWSRetry.exponential_backoff()
def describe_vpc_attr_with_backoff(connection, vpc_id, vpc_attribute):
    """
    Describe VPC Attributes with AWSRetry backoff throttling support.

    connection  : boto3 client connection object
    vpc_id      : The VPC ID to pull attribute value from
    vpc_attribute     : The VPC attribute to get the value from - valid options = enableDnsSupport or enableDnsHostnames
    """

    return connection.describe_vpc_attribute(VpcId=vpc_id, Attribute=vpc_attribute)


def describe_vpcs(connection, module):
    """
    Describe VPCs.

    connection  : boto3 client connection object
    module  : AnsibleModule object
    """
    # collect parameters
    filters = ansible_dict_to_boto3_filter_list(module.params.get('filters'))
    vpc_ids = module.params.get('vpc_ids')

    # init empty list for return vars
    vpc_info = list()
    vpc_list = list()

    # Get the basic VPC info
    try:
        response = connection.describe_vpcs(VpcIds=vpc_ids, Filters=filters)
    except botocore.exceptions.ClientError as e:
        module.fail_json(msg="Unable to describe VPCs {0}: {1}".format(vpc_ids, to_native(e)),
                         exception=traceback.format_exc(), **camel_dict_to_snake_dict(e.response))
    except botocore.exceptions.BotoCoreError as e:
        module.fail_json(msg="Unable to describe VPCs {0}: {1}".format(vpc_ids, to_native(e)),
                         exception=traceback.format_exc())

    # Loop through results and create a list of VPC IDs
    for vpc in response['Vpcs']:
        vpc_list.append(vpc['VpcId'])

    # We can get these results in bulk but still needs two separate calls to the API
    try:
        cl_enabled = connection.describe_vpc_classic_link(VpcIds=vpc_list)
    except botocore.exceptions.ClientError as e:
        if e.response["Error"]["Message"] == "The functionality you requested is not available in this region.":
            cl_enabled = {'Vpcs': [{'VpcId': vpc_id, 'ClassicLinkEnabled': False} for vpc_id in vpc_list]}
        else:
            module.fail_json(msg="Unable to describe if ClassicLink is enabled: {0}".format(to_native(e)),
                             exception=traceback.format_exc(), **camel_dict_to_snake_dict(e.response))
    except botocore.exceptions.BotoCoreError as e:
        module.fail_json(msg="Unable to describe if ClassicLink is enabled: {0}".format(to_native(e)),
                         exception=traceback.format_exc())

    try:
        cl_dns_support = connection.describe_vpc_classic_link_dns_support(VpcIds=vpc_list)
    except botocore.exceptions.ClientError as e:
        if e.response["Error"]["Message"] == "The functionality you requested is not available in this region.":
            cl_dns_support = {'Vpcs': [{'VpcId': vpc_id, 'ClassicLinkDnsSupported': False} for vpc_id in vpc_list]}
        else:
            module.fail_json(msg="Unable to describe if ClassicLinkDns is supported: {0}".format(to_native(e)),
                             exception=traceback.format_exc(), **camel_dict_to_snake_dict(e.response))
    except botocore.exceptions.BotoCoreError as e:
        module.fail_json(msg="Unable to describe if ClassicLinkDns is supported: {0}".format(to_native(e)),
                         exception=traceback.format_exc())

    # Loop through the results and add the other VPC attributes we gathered
    for vpc in response['Vpcs']:
        error_message = "Unable to describe VPC attribute {0}: {1}"
        # We have to make two separate calls per VPC to get these attributes.
        try:
            dns_support = describe_vpc_attr_with_backoff(connection, vpc['VpcId'], 'enableDnsSupport')
        except botocore.exceptions.ClientError as e:
            module.fail_json(msg=error_message.format('enableDnsSupport', to_native(e)),
                             exception=traceback.format_exc(), **camel_dict_to_snake_dict(e.response))
        except botocore.exceptions.BotoCoreError as e:
            module.fail_json(msg=error_message.format('enableDnsSupport', to_native(e)),
                             exception=traceback.format_exc())
        try:
            dns_hostnames = describe_vpc_attr_with_backoff(connection, vpc['VpcId'], 'enableDnsHostnames')
        except botocore.exceptions.ClientError as e:
            module.fail_json(msg=error_message.format('enableDnsHostnames', to_native(e)),
                             exception=traceback.format_exc(), **camel_dict_to_snake_dict(e.response))
        except botocore.exceptions.BotoCoreError as e:
            module.fail_json(msg=error_message.format('enableDnsHostnames', to_native(e)),
                             exception=traceback.format_exc())

        # loop through the ClassicLink Enabled results and add the value for the correct VPC
        for item in cl_enabled['Vpcs']:
            if vpc['VpcId'] == item['VpcId']:
                vpc['ClassicLinkEnabled'] = item['ClassicLinkEnabled']

        # loop through the ClassicLink DNS support results and add the value for the correct VPC
        for item in cl_dns_support['Vpcs']:
            if vpc['VpcId'] == item['VpcId']:
                vpc['ClassicLinkDnsSupported'] = item['ClassicLinkDnsSupported']

        # add the two DNS attributes
        vpc['EnableDnsSupport'] = dns_support['EnableDnsSupport'].get('Value')
        vpc['EnableDnsHostnames'] = dns_hostnames['EnableDnsHostnames'].get('Value')
        # for backwards compatibility
        vpc['id'] = vpc['VpcId']
        vpc_info.append(camel_dict_to_snake_dict(vpc))
        # convert tag list to ansible dict
        vpc_info[-1]['tags'] = boto3_tag_list_to_ansible_dict(vpc.get('Tags', []))

    module.exit_json(vpcs=vpc_info)


def main():
    argument_spec = ec2_argument_spec()
    argument_spec.update(dict(
        vpc_ids=dict(type='list', default=[]),
        filters=dict(type='dict', default={})
    ))

    module = AnsibleModule(argument_spec=argument_spec, supports_check_mode=True)

    if not HAS_BOTO3:
        module.fail_json(msg='boto3 and botocore are required for this module')

    region, ec2_url, aws_connect_params = get_aws_connection_info(module, boto3=True)
    connection = boto3_conn(module, conn_type='client', resource='ec2', region=region, endpoint=ec2_url, **aws_connect_params)

    describe_vpcs(connection, module)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""Show python and pip versions."""

import os
import sys

try:
    import pip
except ImportError:
    pip = None

print(sys.version)

if pip:
    print('pip %s from %s' % (pip.__version__, os.path.dirname(pip.__file__)))



#!/usr/bin/python

# Copyright: (c) 2018, Aaron Smith <ajsmith10381@gmail.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}


DOCUMENTATION = r'''
---
module: aws_config_aggregator
short_description: Manage AWS Config aggregations across multiple accounts
description:
    - Module manages AWS Config resources
version_added: "2.6"
requirements: [ 'botocore', 'boto3' ]
author:
    - "Aaron Smith (@slapula)"
options:
  name:
    description:
    - The name of the AWS Config resource.
    required: true
  state:
    description:
    - Whether the Config rule should be present or absent.
    default: present
    choices: ['present', 'absent']
  account_sources:
    description:
    - Provides a list of source accounts and regions to be aggregated.
    suboptions:
      account_ids:
        description:
        - A list of 12-digit account IDs of accounts being aggregated.
      aws_regions:
        description:
        - A list of source regions being aggregated.
      all_aws_regions:
        description:
        - If true, aggreagate existing AWS Config regions and future regions.
  organization_source:
    description:
    - The region authorized to collect aggregated data.
    suboptions:
      role_arn:
        description:
        - ARN of the IAM role used to retreive AWS Organization details associated with the aggregator account.
      aws_regions:
        description:
        - The source regions being aggregated.
      all_aws_regions:
        description:
        - If true, aggreagate existing AWS Config regions and future regions.
extends_documentation_fragment:
  - aws
  - ec2
'''

EXAMPLES = r'''
- name: Create cross-account aggregator
  aws_config_aggregator:
    name: test_config_rule
    state: present
    account_sources:
      account_ids:
      - 1234567890
      - 0123456789
      - 9012345678
      all_aws_regions: yes
'''

RETURN = r'''#'''


try:
    import botocore
    from botocore.exceptions import BotoCoreError, ClientError
except ImportError:
    pass  # handled by AnsibleAWSModule

from ansible.module_utils.aws.core import AnsibleAWSModule
from ansible.module_utils.ec2 import boto3_conn, get_aws_connection_info, AWSRetry
from ansible.module_utils.ec2 import camel_dict_to_snake_dict, boto3_tag_list_to_ansible_dict


def resource_exists(client, module, resource_type, params):
    try:
        aggregator = client.describe_configuration_aggregators(
            ConfigurationAggregatorNames=[params['name']]
        )
        return aggregator['ConfigurationAggregators'][0]
    except client.exceptions.from_code('NoSuchConfigurationAggregatorException'):
        return
    except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
        module.fail_json_aws(e)


def create_resource(client, module, params, result):
    try:
        response = client.put_configuration_aggregator(
            ConfigurationAggregatorName=params['ConfigurationAggregatorName'],
            AccountAggregationSources=params['AccountAggregationSources'],
            OrganizationAggregationSource=params['OrganizationAggregationSource']
        )
        result['changed'] = True
        result['aggregator'] = camel_dict_to_snake_dict(resource_exists(client, module, params))
        return result
    except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
        module.fail_json_aws(e, msg="Couldn't create AWS Config configuration aggregator")


def update_resource(client, module, resource_type, params, result):
    current_params = client.describe_configuration_aggregators(
        ConfigurationAggregatorNames=[params['name']]
    )

    del current_params['ConfigurationAggregatorArn']
    del current_params['CreationTime']
    del current_params['LastUpdatedTime']

    if params != current_params['ConfigurationAggregators'][0]:
        try:
            client.put_configuration_aggregator(
                ConfigurationAggregatorName=params['ConfigurationAggregatorName'],
                AccountAggregationSources=params['AccountAggregationSources'],
                OrganizationAggregationSource=params['OrganizationAggregationSource']
            )
            result['changed'] = True
            result['aggregator'] = camel_dict_to_snake_dict(resource_exists(client, module, params))
            return result
        except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
            module.fail_json_aws(e, msg="Couldn't create AWS Config configuration aggregator")


def delete_resource(client, module, resource_type, params, result):
    try:
        client.delete_configuration_aggregator(
            ConfigurationAggregatorName=params['ConfigurationAggregatorName']
        )
        result['changed'] = True
        return result
    except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
        module.fail_json_aws(e, msg="Couldn't delete AWS Config configuration aggregator")


def main():
    module = AnsibleAWSModule(
        argument_spec={
            'name': dict(type='str', required=True),
            'state': dict(type='str', choices=['present', 'absent'], default='present'),
            'account_sources': dict(type='list', required=True),
            'organization_source': dict(type='dict', required=True)
        },
        supports_check_mode=False,
    )

    result = {
        'changed': False
    }

    name = module.params.get('name')
    state = module.params.get('state')

    params = {}
    if name:
        params['ConfigurationAggregatorName'] = name
    if module.params.get('account_sources'):
        params['AccountAggregationSources'] = []
        for i in module.params.get('account_sources'):
            tmp_dict = {}
            if i.get('account_ids'):
                tmp_dict['AccountIds'] = i.get('account_ids')
            if i.get('aws_regions'):
                tmp_dict['AwsRegions'] = i.get('aws_regions')
            if i.get('all_aws_regions') is not None:
                tmp_dict['AllAwsRegions'] = i.get('all_aws_regions')
            params['AccountAggregationSources'].append(tmp_dict)
    if module.params.get('organization_source'):
        params['OrganizationAggregationSource'] = {}
        if module.params.get('organization_source').get('role_arn'):
            params['OrganizationAggregationSource'].update({
                'RoleArn': module.params.get('organization_source').get('role_arn')
            })
        if module.params.get('organization_source').get('aws_regions'):
            params['OrganizationAggregationSource'].update({
                'AwsRegions': module.params.get('organization_source').get('aws_regions')
            })
        if module.params.get('organization_source').get('all_aws_regions') is not None:
            params['OrganizationAggregationSourcep'].update({
                'AllAwsRegions': module.params.get('organization_source').get('all_aws_regions')
            })

    client = module.client('config', retry_decorator=AWSRetry.jittered_backoff())

    resource_status = resource_exists(client, module, params)

    if state == 'present':
        if not resource_status:
            create_resource(client, module, params, result)
        else:
            update_resource(client, module, params, result)

    if state == 'absent':
        if resource_status:
            delete_resource(client, module, params, result)

    module.exit_json(changed=result['changed'])


if __name__ == '__main__':
    main()



#!/usr/bin/python

# (c) 2013, Paul Durivage <paul.durivage@rackspace.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}


DOCUMENTATION = '''
---
module: rax_files_objects
short_description: Upload, download, and delete objects in Rackspace Cloud Files
description:
  - Upload, download, and delete objects in Rackspace Cloud Files
version_added: "1.5"
options:
  clear_meta:
    description:
      - Optionally clear existing metadata when applying metadata to existing objects.
        Selecting this option is only appropriate when setting type=meta
    type: bool
    default: 'no'
  container:
    description:
      - The container to use for file object operations.
    required: true
  dest:
    description:
      - The destination of a "get" operation; i.e. a local directory, "/home/user/myfolder".
        Used to specify the destination of an operation on a remote object; i.e. a file name,
        "file1", or a comma-separated list of remote objects, "file1,file2,file17"
  expires:
    description:
      - Used to set an expiration on a file or folder uploaded to Cloud Files.
        Requires an integer, specifying expiration in seconds
  meta:
    description:
      - A hash of items to set as metadata values on an uploaded file or folder
  method:
    description:
      - The method of operation to be performed.  For example, put to upload files
        to Cloud Files, get to download files from Cloud Files or delete to delete
        remote objects in Cloud Files
    choices:
      - get
      - put
      - delete
    default: get
  src:
    description:
      - Source from which to upload files.  Used to specify a remote object as a source for
        an operation, i.e. a file name, "file1", or a comma-separated list of remote objects,
        "file1,file2,file17".  src and dest are mutually exclusive on remote-only object operations
  structure:
    description:
      - Used to specify whether to maintain nested directory structure when downloading objects
        from Cloud Files.  Setting to false downloads the contents of a container to a single,
        flat directory
    type: bool
    default: 'yes'
  state:
    description:
      - Indicate desired state of the resource
    choices: ['present', 'absent']
    default: present
  type:
    description:
      - Type of object to do work on
      - Metadata object or a file object
    choices:
      - file
      - meta
    default: file
author: "Paul Durivage (@angstwad)"
extends_documentation_fragment:
  - rackspace
  - rackspace.openstack
'''

EXAMPLES = '''
- name: "Test Cloud Files Objects"
  hosts: local
  gather_facts: False
  tasks:
    - name: "Get objects from test container"
      rax_files_objects:
        container: testcont
        dest: ~/Downloads/testcont

    - name: "Get single object from test container"
      rax_files_objects:
        container: testcont
        src: file1
        dest: ~/Downloads/testcont

    - name: "Get several objects from test container"
      rax_files_objects:
        container: testcont
        src: file1,file2,file3
        dest: ~/Downloads/testcont

    - name: "Delete one object in test container"
      rax_files_objects:
        container: testcont
        method: delete
        dest: file1

    - name: "Delete several objects in test container"
      rax_files_objects:
        container: testcont
        method: delete
        dest: file2,file3,file4

    - name: "Delete all objects in test container"
      rax_files_objects:
        container: testcont
        method: delete

    - name: "Upload all files to test container"
      rax_files_objects:
        container: testcont
        method: put
        src: ~/Downloads/onehundred

    - name: "Upload one file to test container"
      rax_files_objects:
        container: testcont
        method: put
        src: ~/Downloads/testcont/file1

    - name: "Upload one file to test container with metadata"
      rax_files_objects:
        container: testcont
        src: ~/Downloads/testcont/file2
        method: put
        meta:
          testkey: testdata
          who_uploaded_this: someuser@example.com

    - name: "Upload one file to test container with TTL of 60 seconds"
      rax_files_objects:
        container: testcont
        method: put
        src: ~/Downloads/testcont/file3
        expires: 60

    - name: "Attempt to get remote object that does not exist"
      rax_files_objects:
        container: testcont
        method: get
        src: FileThatDoesNotExist.jpg
        dest: ~/Downloads/testcont
      ignore_errors: yes

    - name: "Attempt to delete remote object that does not exist"
      rax_files_objects:
        container: testcont
        method: delete
        dest: FileThatDoesNotExist.jpg
      ignore_errors: yes

- name: "Test Cloud Files Objects Metadata"
  hosts: local
  gather_facts: false
  tasks:
    - name: "Get metadata on one object"
      rax_files_objects:
        container: testcont
        type: meta
        dest: file2

    - name: "Get metadata on several objects"
      rax_files_objects:
        container: testcont
        type: meta
        src: file2,file1

    - name: "Set metadata on an object"
      rax_files_objects:
        container: testcont
        type: meta
        dest: file17
        method: put
        meta:
          key1: value1
          key2: value2
        clear_meta: true

    - name: "Verify metadata is set"
      rax_files_objects:
        container: testcont
        type: meta
        src: file17

    - name: "Delete metadata"
      rax_files_objects:
        container: testcont
        type: meta
        dest: file17
        method: delete
        meta:
          key1: ''
          key2: ''

    - name: "Get metadata on all objects"
      rax_files_objects:
        container: testcont
        type: meta
'''

import os

try:
    import pyrax
    HAS_PYRAX = True
except ImportError:
    HAS_PYRAX = False

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.rax import rax_argument_spec, rax_required_together, setup_rax_module


EXIT_DICT = dict(success=False)
META_PREFIX = 'x-object-meta-'


def _get_container(module, cf, container):
    try:
        return cf.get_container(container)
    except pyrax.exc.NoSuchContainer as e:
        module.fail_json(msg=e.message)


def _upload_folder(cf, folder, container, ttl=None, headers=None):
    """ Uploads a folder to Cloud Files.
    """
    total_bytes = 0
    for root, dirs, files in os.walk(folder):
        for fname in files:
            full_path = os.path.join(root, fname)
            obj_name = os.path.relpath(full_path, folder)
            obj_size = os.path.getsize(full_path)
            cf.upload_file(container, full_path,
                           obj_name=obj_name, return_none=True, ttl=ttl, headers=headers)
            total_bytes += obj_size
    return total_bytes


def upload(module, cf, container, src, dest, meta, expires):
    """ Uploads a single object or a folder to Cloud Files Optionally sets an
    metadata, TTL value (expires), or Content-Disposition and Content-Encoding
    headers.
    """
    if not src:
        module.fail_json(msg='src must be specified when uploading')

    c = _get_container(module, cf, container)
    src = os.path.abspath(os.path.expanduser(src))
    is_dir = os.path.isdir(src)

    if not is_dir and not os.path.isfile(src) or not os.path.exists(src):
        module.fail_json(msg='src must be a file or a directory')
    if dest and is_dir:
        module.fail_json(msg='dest cannot be set when whole '
                             'directories are uploaded')

    cont_obj = None
    total_bytes = 0
    if dest and not is_dir:
        try:
            cont_obj = c.upload_file(src, obj_name=dest, ttl=expires, headers=meta)
        except Exception as e:
            module.fail_json(msg=e.message)
    elif is_dir:
        try:
            total_bytes = _upload_folder(cf, src, c, ttl=expires, headers=meta)
        except Exception as e:
            module.fail_json(msg=e.message)
    else:
        try:
            cont_obj = c.upload_file(src, ttl=expires, headers=meta)
        except Exception as e:
            module.fail_json(msg=e.message)

    EXIT_DICT['success'] = True
    EXIT_DICT['container'] = c.name
    EXIT_DICT['msg'] = "Uploaded %s to container: %s" % (src, c.name)
    if cont_obj or total_bytes > 0:
        EXIT_DICT['changed'] = True
    if meta:
        EXIT_DICT['meta'] = dict(updated=True)

    if cont_obj:
        EXIT_DICT['bytes'] = cont_obj.total_bytes
        EXIT_DICT['etag'] = cont_obj.etag
    else:
        EXIT_DICT['bytes'] = total_bytes

    module.exit_json(**EXIT_DICT)


def download(module, cf, container, src, dest, structure):
    """ Download objects from Cloud Files to a local path specified by "dest".
    Optionally disable maintaining a directory structure by by passing a
    false value to "structure".
    """
    # Looking for an explicit destination
    if not dest:
        module.fail_json(msg='dest is a required argument when '
                             'downloading from Cloud Files')

    # Attempt to fetch the container by name
    c = _get_container(module, cf, container)

    # Accept a single object name or a comma-separated list of objs
    # If not specified, get the entire container
    if src:
        objs = src.split(',')
        objs = map(str.strip, objs)
    else:
        objs = c.get_object_names()

    dest = os.path.abspath(os.path.expanduser(dest))
    is_dir = os.path.isdir(dest)

    if not is_dir:
        module.fail_json(msg='dest must be a directory')

    results = []
    for obj in objs:
        try:
            c.download_object(obj, dest, structure=structure)
        except Exception as e:
            module.fail_json(msg=e.message)
        else:
            results.append(obj)

    len_results = len(results)
    len_objs = len(objs)

    EXIT_DICT['container'] = c.name
    EXIT_DICT['requested_downloaded'] = results
    if results:
        EXIT_DICT['changed'] = True
    if len_results == len_objs:
        EXIT_DICT['success'] = True
        EXIT_DICT['msg'] = "%s objects downloaded to %s" % (len_results, dest)
    else:
        EXIT_DICT['msg'] = "Error: only %s of %s objects were " \
                           "downloaded" % (len_results, len_objs)
    module.exit_json(**EXIT_DICT)


def delete(module, cf, container, src, dest):
    """ Delete specific objects by proving a single file name or a
    comma-separated list to src OR dest (but not both).  Omitting file name(s)
    assumes the entire container is to be deleted.
    """
    objs = None
    if src and dest:
        module.fail_json(msg="Error: ambiguous instructions; files to be deleted "
                             "have been specified on both src and dest args")
    elif dest:
        objs = dest
    else:
        objs = src

    c = _get_container(module, cf, container)

    if objs:
        objs = objs.split(',')
        objs = map(str.strip, objs)
    else:
        objs = c.get_object_names()

    num_objs = len(objs)

    results = []
    for obj in objs:
        try:
            result = c.delete_object(obj)
        except Exception as e:
            module.fail_json(msg=e.message)
        else:
            results.append(result)

    num_deleted = results.count(True)

    EXIT_DICT['container'] = c.name
    EXIT_DICT['deleted'] = num_deleted
    EXIT_DICT['requested_deleted'] = objs

    if num_deleted:
        EXIT_DICT['changed'] = True

    if num_objs == num_deleted:
        EXIT_DICT['success'] = True
        EXIT_DICT['msg'] = "%s objects deleted" % num_deleted
    else:
        EXIT_DICT['msg'] = ("Error: only %s of %s objects "
                            "deleted" % (num_deleted, num_objs))
    module.exit_json(**EXIT_DICT)


def get_meta(module, cf, container, src, dest):
    """ Get metadata for a single file, comma-separated list, or entire
    container
    """
    c = _get_container(module, cf, container)

    objs = None
    if src and dest:
        module.fail_json(msg="Error: ambiguous instructions; files to be deleted "
                             "have been specified on both src and dest args")
    elif dest:
        objs = dest
    else:
        objs = src

    if objs:
        objs = objs.split(',')
        objs = map(str.strip, objs)
    else:
        objs = c.get_object_names()

    results = dict()
    for obj in objs:
        try:
            meta = c.get_object(obj).get_metadata()
        except Exception as e:
            module.fail_json(msg=e.message)
        else:
            results[obj] = dict()
            for k, v in meta.items():
                meta_key = k.split(META_PREFIX)[-1]
                results[obj][meta_key] = v

    EXIT_DICT['container'] = c.name
    if results:
        EXIT_DICT['meta_results'] = results
        EXIT_DICT['success'] = True
    module.exit_json(**EXIT_DICT)


def put_meta(module, cf, container, src, dest, meta, clear_meta):
    """ Set metadata on a container, single file, or comma-separated list.
    Passing a true value to clear_meta clears the metadata stored in Cloud
    Files before setting the new metadata to the value of "meta".
    """
    objs = None
    if src and dest:
        module.fail_json(msg="Error: ambiguous instructions; files to set meta"
                             " have been specified on both src and dest args")
    elif dest:
        objs = dest
    else:
        objs = src

    objs = objs.split(',')
    objs = map(str.strip, objs)

    c = _get_container(module, cf, container)

    results = []
    for obj in objs:
        try:
            result = c.get_object(obj).set_metadata(meta, clear=clear_meta)
        except Exception as e:
            module.fail_json(msg=e.message)
        else:
            results.append(result)

    EXIT_DICT['container'] = c.name
    EXIT_DICT['success'] = True
    if results:
        EXIT_DICT['changed'] = True
        EXIT_DICT['num_changed'] = True
    module.exit_json(**EXIT_DICT)


def delete_meta(module, cf, container, src, dest, meta):
    """ Removes metadata keys and values specified in meta, if any.  Deletes on
    all objects specified by src or dest (but not both), if any; otherwise it
    deletes keys on all objects in the container
    """
    objs = None
    if src and dest:
        module.fail_json(msg="Error: ambiguous instructions; meta keys to be "
                             "deleted have been specified on both src and dest"
                             " args")
    elif dest:
        objs = dest
    else:
        objs = src

    objs = objs.split(',')
    objs = map(str.strip, objs)

    c = _get_container(module, cf, container)

    results = []  # Num of metadata keys removed, not objects affected
    for obj in objs:
        if meta:
            for k, v in meta.items():
                try:
                    result = c.get_object(obj).remove_metadata_key(k)
                except Exception as e:
                    module.fail_json(msg=e.message)
                else:
                    results.append(result)
        else:
            try:
                o = c.get_object(obj)
            except pyrax.exc.NoSuchObject as e:
                module.fail_json(msg=e.message)

            for k, v in o.get_metadata().items():
                try:
                    result = o.remove_metadata_key(k)
                except Exception as e:
                    module.fail_json(msg=e.message)
                results.append(result)

    EXIT_DICT['container'] = c.name
    EXIT_DICT['success'] = True
    if results:
        EXIT_DICT['changed'] = True
        EXIT_DICT['num_deleted'] = len(results)
    module.exit_json(**EXIT_DICT)


def cloudfiles(module, container, src, dest, method, typ, meta, clear_meta,
               structure, expires):
    """ Dispatch from here to work with metadata or file objects """
    cf = pyrax.cloudfiles

    if cf is None:
        module.fail_json(msg='Failed to instantiate client. This '
                             'typically indicates an invalid region or an '
                             'incorrectly capitalized region name.')

    if typ == "file":
        if method == 'put':
            upload(module, cf, container, src, dest, meta, expires)

        elif method == 'get':
            download(module, cf, container, src, dest, structure)

        elif method == 'delete':
            delete(module, cf, container, src, dest)

    else:
        if method == 'get':
            get_meta(module, cf, container, src, dest)

        if method == 'put':
            put_meta(module, cf, container, src, dest, meta, clear_meta)

        if method == 'delete':
            delete_meta(module, cf, container, src, dest, meta)


def main():
    argument_spec = rax_argument_spec()
    argument_spec.update(
        dict(
            container=dict(required=True),
            src=dict(),
            dest=dict(),
            method=dict(default='get', choices=['put', 'get', 'delete']),
            type=dict(default='file', choices=['file', 'meta']),
            meta=dict(type='dict', default=dict()),
            clear_meta=dict(default=False, type='bool'),
            structure=dict(default=True, type='bool'),
            expires=dict(type='int'),
        )
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        required_together=rax_required_together()
    )

    if not HAS_PYRAX:
        module.fail_json(msg='pyrax is required for this module')

    container = module.params.get('container')
    src = module.params.get('src')
    dest = module.params.get('dest')
    method = module.params.get('method')
    typ = module.params.get('type')
    meta = module.params.get('meta')
    clear_meta = module.params.get('clear_meta')
    structure = module.params.get('structure')
    expires = module.params.get('expires')

    if clear_meta and not typ == 'meta':
        module.fail_json(msg='clear_meta can only be used when setting metadata')

    setup_rax_module(module, pyrax)
    cloudfiles(module, container, src, dest, method, typ, meta, clear_meta, structure, expires)


if __name__ == '__main__':
    main()



# This is a standalone test for the regex inside validate-modules
# It is not suitable to add to the make tests target because the
# file under test is outside the test's sys.path AND has a hyphen
# in the name making it unimportable.
#
# To execute this by hand:
#   1) cd <checkoutdir>
#   2) source hacking/env-setup
#   3) PYTHONPATH=./lib nosetests -d -w test -v --nocapture sanity/validate-modules

import re
from ansible.compat.tests import unittest

# TYPE_REGEX = re.compile(r'.*\stype\(.*')
# TYPE_REGEX = re.compile(r'.*(if|or)\stype\(.*')
# TYPE_REGEX = re.compile(r'.*(if|or)(\s+.*|\s+)type\(.*')
# TYPE_REGEX = re.compile(r'.*(if|or)(\s+.*|\s+)type\(.*')
# TYPE_REGEX = re.compile(r'.*(if|\sor)(\s+.*|\s+)type\(.*')
# TYPE_REGEX =  re.compile(r'.*(if|\sor)(\s+.*|\s+)(?<!_)type\(.*')
TYPE_REGEX = re.compile(r'.*(if|or)(\s+.*|\s+)(?<!_)(?<!str\()type\(.*')


class TestValidateModulesRegex(unittest.TestCase):

    def test_type_regex(self):
        # each of these examples needs to be matched or not matched
        checks = [
            ['if type(foo) is Bar', True],
            ['if Bar is type(foo)', True],
            ['if type(foo) is not Bar', True],
            ['if Bar is not type(foo)', True],
            ['if type(foo) == Bar', True],
            ['if Bar == type(foo)', True],
            ['if type(foo)==Bar', True],
            ['if Bar==type(foo)', True],
            ['if type(foo) != Bar', True],
            ['if Bar != type(foo)', True],
            ['if type(foo)!=Bar', True],
            ['if Bar!=type(foo)', True],
            ['if foo or type(bar) != Bar', True],
            ['x = type(foo)', False],
            ["error = err.message + ' ' + str(err) + ' - ' + str(type(err))", False],
            # cloud/amazon/ec2_group.py
            ["module.fail_json(msg='Invalid rule parameter type [%s].' % type(rule))", False],
            # files/patch.py
            ["p = type('Params', (), module.params)", False],  # files/patch.py
            # system/osx_defaults.py
            ["if self.current_value is not None and not isinstance(self.current_value, type(self.value)):", True],
            # system/osx_defaults.py
            ['raise OSXDefaultsException("Type mismatch. Type in defaults: " + type(self.current_value).__name__)', False],
            # network/nxos/nxos_interface.py
            ["if get_interface_type(interface) == 'svi':", False],
        ]
        for idc, check in enumerate(checks):
            cstring = check[0]
            cexpected = check[1]

            match = TYPE_REGEX.match(cstring)
            if cexpected and not match:
                assert False, "%s should have matched" % cstring
            elif not cexpected and match:
                assert False, "%s should not have matched" % cstring



#!/usr/bin/python
# -*- coding: utf-8 -*-

# (c) 2015, Matt Makai <matthew.makai@gmail.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}


DOCUMENTATION = '''
---
version_added: "1.6"
module: twilio
short_description: Sends a text message to a mobile phone through Twilio.
description:
   - Sends a text message to a phone number through the Twilio messaging API.
notes:
   - This module is non-idempotent because it sends an email through the
     external API. It is idempotent only in the case that the module fails.
   - Like the other notification modules, this one requires an external
     dependency to work. In this case, you'll need a Twilio account with
     a purchased or verified phone number to send the text message.
options:
  account_sid:
    description:
      user's Twilio account token found on the account page
    required: true
  auth_token:
    description: user's Twilio authentication token
    required: true
  msg:
    description:
      the body of the text message
    required: true
  to_number:
    description:
      one or more phone numbers to send the text message to,
      format +15551112222
    required: true
  from_number:
    description:
      the Twilio number to send the text message from, format +15551112222
    required: true
  media_url:
    description:
      a URL with a picture, video or sound clip to send with an MMS
      (multimedia message) instead of a plain SMS
    required: false

author: "Matt Makai (@makaimc)"
'''

EXAMPLES = '''
# send an SMS about the build status to (555) 303 5681
# note: replace account_sid and auth_token values with your credentials
# and you have to have the 'from_number' on your Twilio account
- twilio:
    msg: All servers with webserver role are now configured.
    account_sid: ACXXXXXXXXXXXXXXXXX
    auth_token: ACXXXXXXXXXXXXXXXXX
    from_number: +15552014545
    to_number: +15553035681
  delegate_to: localhost

# send an SMS to multiple phone numbers about the deployment
# note: replace account_sid and auth_token values with your credentials
# and you have to have the 'from_number' on your Twilio account
- twilio:
    msg: This server configuration is now complete.
    account_sid: ACXXXXXXXXXXXXXXXXX
    auth_token: ACXXXXXXXXXXXXXXXXX
    from_number: +15553258899
    to_number:
      - +15551113232
      - +12025551235
      - +19735559010
  delegate_to: localhost

# send an MMS to a single recipient with an update on the deployment
# and an image of the results
# note: replace account_sid and auth_token values with your credentials
# and you have to have the 'from_number' on your Twilio account
- twilio:
    msg: Deployment complete!
    account_sid: ACXXXXXXXXXXXXXXXXX
    auth_token: ACXXXXXXXXXXXXXXXXX
    from_number: +15552014545
    to_number: +15553035681
    media_url: https://demo.twilio.com/logo.png
  delegate_to: localhost
'''

# =======================================
# twilio module support methods
#
import json

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.six.moves.urllib.parse import urlencode
from ansible.module_utils.urls import fetch_url


def post_twilio_api(module, account_sid, auth_token, msg, from_number,
                    to_number, media_url=None):
    URI = "https://api.twilio.com/2010-04-01/Accounts/%s/Messages.json" \
        % (account_sid,)
    AGENT = "Ansible"

    data = {'From': from_number, 'To': to_number, 'Body': msg}
    if media_url:
        data['MediaUrl'] = media_url
    encoded_data = urlencode(data)

    headers = {'User-Agent': AGENT,
               'Content-type': 'application/x-www-form-urlencoded',
               'Accept': 'application/json',
               }

    # Hack module params to have the Basic auth params that fetch_url expects
    module.params['url_username'] = account_sid.replace('\n', '')
    module.params['url_password'] = auth_token.replace('\n', '')

    return fetch_url(module, URI, data=encoded_data, headers=headers)


# =======================================
# Main
#

def main():

    module = AnsibleModule(
        argument_spec=dict(
            account_sid=dict(required=True),
            auth_token=dict(required=True, no_log=True),
            msg=dict(required=True),
            from_number=dict(required=True),
            to_number=dict(required=True),
            media_url=dict(default=None, required=False),
        ),
        supports_check_mode=True
    )

    account_sid = module.params['account_sid']
    auth_token = module.params['auth_token']
    msg = module.params['msg']
    from_number = module.params['from_number']
    to_number = module.params['to_number']
    media_url = module.params['media_url']

    if not isinstance(to_number, list):
        to_number = [to_number]

    for number in to_number:
        r, info = post_twilio_api(module, account_sid, auth_token, msg,
                                  from_number, number, media_url)
        if info['status'] not in [200, 201]:
            body_message = "unknown error"
            if 'body' in info:
                body = json.loads(info['body'])
                body_message = body['message']
            module.fail_json(msg="unable to send message to %s: %s" % (number, body_message))

    module.exit_json(msg=msg, changed=False)


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2015, Joseph Callen <jcallen () csc.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = r'''
---
module: vmware_cluster
short_description: Manage VMware vSphere clusters
description:
    - Add or remove VMware vSphere clusters.
version_added: '2.0'
author:
- Joseph Callen (@jcpowermac)
requirements:
    - Tested on ESXi 5.5
    - PyVmomi installed
options:
    cluster_name:
        description:
            - The name of the cluster that will be created.
        required: yes
    datacenter_name:
        description:
            - The name of the datacenter the cluster will be created in.
        required: yes
    enable_drs:
        description:
            - If set to C(yes) will enable DRS when the cluster is created.
        type: bool
        default: 'no'
    enable_ha:
        description:
            - If set to C(yes) will enable HA when the cluster is created.
        type: bool
        default: 'no'
    enable_vsan:
        description:
            - If set to C(yes) will enable vSAN when the cluster is created.
        type: bool
        default: 'no'
    state:
        description:
            - Create (C(present)) or remove (C(absent)) a VMware vSphere cluster.
        choices: [absent, present]
        default: present
extends_documentation_fragment: vmware.documentation
'''

EXAMPLES = r'''
- name: Create Cluster
  local_action:
    module: vmware_cluster
    hostname: '{{ ansible_ssh_host }}'
    username: root
    password: vmware
    datacenter_name: datacenter
    cluster_name: cluster
    enable_ha: yes
    enable_drs: yes
    enable_vsan: yes
'''

try:
    from pyVmomi import vim, vmodl
    HAS_PYVMOMI = True
except ImportError:
    HAS_PYVMOMI = False

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.vmware import (HAS_PYVMOMI,
                                         TaskError,
                                         connect_to_api,
                                         find_cluster_by_name_datacenter,
                                         find_datacenter_by_name,
                                         vmware_argument_spec,
                                         wait_for_task
                                         )


class VMwareCluster(object):
    def __init__(self, module):
        self.module = module
        self.cluster_name = module.params['cluster_name']
        self.datacenter_name = module.params['datacenter_name']
        self.enable_drs = module.params['enable_drs']
        self.enable_ha = module.params['enable_ha']
        self.enable_vsan = module.params['enable_vsan']
        self.desired_state = module.params['state']
        self.datacenter = None
        self.cluster = None
        self.content = connect_to_api(module)

    def process_state(self):
        cluster_states = {
            'absent': {
                'present': self.state_destroy_cluster,
                'absent': self.state_exit_unchanged,
            },
            'present': {
                'update': self.state_update_cluster,
                'present': self.state_exit_unchanged,
                'absent': self.state_create_cluster,
            }
        }
        current_state = self.check_cluster_configuration()
        # Based on the desired_state and the current_state call
        # the appropriate method from the dictionary
        cluster_states[self.desired_state][current_state]()

    def configure_ha(self):
        das_config = vim.cluster.DasConfigInfo()
        das_config.enabled = self.enable_ha
        das_config.admissionControlPolicy = vim.cluster.FailoverLevelAdmissionControlPolicy()
        das_config.admissionControlPolicy.failoverLevel = 2
        return das_config

    def configure_drs(self):
        drs_config = vim.cluster.DrsConfigInfo()
        drs_config.enabled = self.enable_drs
        # Set to partially automated
        drs_config.vmotionRate = 3
        return drs_config

    def configure_vsan(self):
        vsan_config = vim.vsan.cluster.ConfigInfo()
        vsan_config.enabled = self.enable_vsan
        vsan_config.defaultConfig = vim.vsan.cluster.ConfigInfo.HostDefaultInfo()
        vsan_config.defaultConfig.autoClaimStorage = False
        return vsan_config

    def state_create_cluster(self):
        try:
            cluster_config_spec = vim.cluster.ConfigSpecEx()
            cluster_config_spec.dasConfig = self.configure_ha()
            cluster_config_spec.drsConfig = self.configure_drs()
            if self.enable_vsan:
                cluster_config_spec.vsanConfig = self.configure_vsan()
            if not self.module.check_mode:
                self.datacenter.hostFolder.CreateClusterEx(self.cluster_name, cluster_config_spec)
            self.module.exit_json(changed=True)
        except vim.fault.DuplicateName:
            self.module.fail_json(msg="A cluster with the name %s already exists" % self.cluster_name)
        except vmodl.fault.InvalidArgument:
            self.module.fail_json(msg="Cluster configuration specification parameter is invalid")
        except vim.fault.InvalidName:
            self.module.fail_json(msg="%s is an invalid name for a cluster" % self.cluster_name)
        except vmodl.fault.NotSupported:
            # This should never happen
            self.module.fail_json(msg="Trying to create a cluster on an incorrect folder object")
        except vmodl.RuntimeFault as runtime_fault:
            self.module.fail_json(msg=runtime_fault.msg)
        except vmodl.MethodFault as method_fault:
            # This should never happen either
            self.module.fail_json(msg=method_fault.msg)

    def state_destroy_cluster(self):
        changed = True
        result = None

        try:
            if not self.module.check_mode:
                task = self.cluster.Destroy_Task()
                changed, result = wait_for_task(task)
            self.module.exit_json(changed=changed, result=result)
        except vim.fault.VimFault as vim_fault:
            self.module.fail_json(msg=vim_fault.msg)
        except vmodl.RuntimeFault as runtime_fault:
            self.module.fail_json(msg=runtime_fault.msg)
        except vmodl.MethodFault as method_fault:
            self.module.fail_json(msg=method_fault.msg)

    def state_exit_unchanged(self):
        self.module.exit_json(changed=False)

    def state_update_cluster(self):
        cluster_config_spec = vim.cluster.ConfigSpecEx()
        changed = True
        result = None

        if self.cluster.configurationEx.dasConfig.enabled != self.enable_ha:
            cluster_config_spec.dasConfig = self.configure_ha()
        if self.cluster.configurationEx.drsConfig.enabled != self.enable_drs:
            cluster_config_spec.drsConfig = self.configure_drs()
        if self.cluster.configurationEx.vsanConfigInfo.enabled != self.enable_vsan:
            cluster_config_spec.vsanConfig = self.configure_vsan()

        try:
            if not self.module.check_mode:
                task = self.cluster.ReconfigureComputeResource_Task(cluster_config_spec, True)
                changed, result = wait_for_task(task)
            self.module.exit_json(changed=changed, result=result)
        except vmodl.RuntimeFault as runtime_fault:
            self.module.fail_json(msg=runtime_fault.msg)
        except vmodl.MethodFault as method_fault:
            self.module.fail_json(msg=method_fault.msg)
        except TaskError as task_e:
            self.module.fail_json(msg=str(task_e))

    def check_cluster_configuration(self):
        try:
            self.datacenter = find_datacenter_by_name(self.content, self.datacenter_name)
            if self.datacenter is None:
                self.module.fail_json(msg="Datacenter %s does not exist, "
                                          "please create first with Ansible Module vmware_datacenter or manually."
                                          % self.datacenter_name)
            self.cluster = find_cluster_by_name_datacenter(self.datacenter, self.cluster_name)

            if self.cluster is None:
                return 'absent'
            else:
                desired_state = (self.enable_ha,
                                 self.enable_drs,
                                 self.enable_vsan)

                current_state = (self.cluster.configurationEx.dasConfig.enabled,
                                 self.cluster.configurationEx.drsConfig.enabled,
                                 self.cluster.configurationEx.vsanConfigInfo.enabled)

                if desired_state != current_state:
                    return 'update'
                else:
                    return 'present'
        except vmodl.RuntimeFault as runtime_fault:
            self.module.fail_json(msg=runtime_fault.msg)
        except vmodl.MethodFault as method_fault:
            self.module.fail_json(msg=method_fault.msg)


def main():

    argument_spec = vmware_argument_spec()
    argument_spec.update(dict(
        cluster_name=dict(type='str', required=True),
        datacenter_name=dict(type='str', required=True),
        enable_drs=dict(type='bool', default=False),
        enable_ha=dict(type='bool', default=False),
        enable_vsan=dict(type='bool', default=False),
        state=dict(type='str', default='present', choices=['absent', 'present']),
    ))

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
    )

    if not HAS_PYVMOMI:
        module.fail_json(msg='pyvmomi is required for this module')

    vmware_cluster = VMwareCluster(module)
    vmware_cluster.process_state()


if __name__ == '__main__':
    main()



# This code is part of Ansible, but is an independent component.
# This particular file snippet, and this file snippet only, is BSD licensed.
# Modules you write using this snippet, which is embedded dynamically by
# Ansible still belong to the author of the module, and may assign their own
# license to the complete work.
#
# Copyright (C) 2017 Lenovo, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Contains error codes and methods
# Lenovo Networking

errorDict = {0: 'Success',
             1: 'NOK',
             101: 'Device Response Timed out',
             102: 'Command Not supported - Use CLI command',
             103: 'Invalid Context',
             104: 'Command Value Not Supported as of Now. Use vlan Id only',
             105: 'Invalid interface Range',
             106: 'Please provide Enable Password.',
             108: '',
             109: '',
             110: 'Invalid protocol option',
             111: 'The Value is not Integer',
             112: 'The Value is not Float',
             113: 'Value is not in Range',
             114: 'Range value is not Integer',
             115: 'Value is not in Options',
             116: 'The Value is not Long',
             117: 'Range value is not Long',
             118: 'The Value cannot be empty',
             119: 'The Value is not String',
             120: 'The Value is not Matching',
             121: 'The Value is not IPV4 Address',
             122: 'The Value is not IPV6 Address',
             123: '',
             124: '',
             125: '',
             126: '',
             127: '',
             128: '',
             129: '',
             130: 'Invalid Access Map Name',
             131: 'Invalid Vlan Dot1q Tag',
             132: 'Invalid Vlan filter value',
             133: 'Invalid Vlan Range Value',
             134: 'Invalid Vlan Id',
             135: 'Invalid Vlan Access Map Action',
             136: 'Invalid Vlan Access Map Name',
             137: 'Invalid Access List',
             138: 'Invalid Vlan Access Map parameter',
             139: 'Invalid Vlan Name',
             140: 'Invalid Vlan Flood value,',
             141: 'Invalid Vlan State Value',
             142: 'Invalid Vlan Last Member query Interval',
             143: 'Invalid Querier IP address',
             144: 'Invalid Querier Time out',
             145: 'Invalid Query Interval',
             146: 'Invalid Vlan query max response time',
             147: 'Invalid vlan robustness variable',
             148: 'Invalid Vlan Startup Query count',
             149: 'Invalid vlan Startup Query Interval',
             150: 'Invalid Vlan snooping version',
             151: 'Invalid Vlan Ethernet Interface',
             152: 'Invalid Vlan Port Tag Number',
             153: 'Invalid mrouter option',
             154: 'Invalid Vlan Option',
             155: '',
             156: '',
             157: '',
             158: '',
             159: '',
             160: 'Invalid Vlag Auto Recovery Value',
             161: 'Invalid Vlag Config Consistency Value',
             162: 'Invalid Vlag Port Aggregation Number',
             163: 'Invalid Vlag Priority Value',
             164: 'Invalid Vlag Startup delay value',
             165: 'Invalid Vlag Trie Id',
             166: 'Invalid Vlag Instance Option',
             167: 'Invalid Vlag Keep Alive Attempts',
             168: 'Invalid Vlag Keep Alive Interval',
             169: 'Invalid Vlag Retry Interval',
             170: 'Invalid Vlag Peer Ip VRF Value',
             171: 'Invalid Vlag Health Check Options',
             172: 'Invalid Vlag Option',
             173: '',
             174: '',
             175: '',
             176: 'Invalid BGP As Number',
             177: 'Invalid Routing protocol option',
             178: 'Invalid BGP Address Family',
             179: 'Invalid AS Path options',
             180: 'Invalid BGP med options',
             181: 'Invalid Best Path option',
             182: 'Invalid BGP Local count number',
             183: 'Cluster Id has to either IP or AS Number',
             184: 'Invalid confederation identifier',
             185: 'Invalid Confederation Peer AS Value',
             186: 'Invalid Confederation Option',
             187: 'Invalid state path relay value',
             188: 'Invalid Maxas Limit AS Value',
             189: 'Invalid Neighbor IP Address or Neighbor AS Number',
             190: 'Invalid Router Id',
             191: 'Invalid BGP Keep Alive Interval',
             192: 'Invalid BGP Hold time',
             193: 'Invalid BGP Option',
             194: 'Invalid BGP Address Family option',
             195: 'Invalid BGP Address Family Redistribution option. ',
             196: 'Invalid BGP Address Family Route Map Name',
             197: 'Invalid Next Hop Critical Delay',
             198: 'Invalid Next Hop Non Critical Delay',
             199: 'Invalid Multipath Number Value',
             200: 'Invalid Aggegation Group Mode',
             201: 'Invalid Aggregation Group No',
             202: 'Invalid BFD Access Vlan',
             203: 'Invalid CFD Bridgeport Mode',
             204: 'Invalid Trunk Option',
             205: 'Invalid BFD Option',
             206: 'Invalid Portchannel description',
             207: 'Invalid Portchannel duplex option',
             208: 'Invalid Flow control option state',
             209: 'Invalid Flow control option',
             210: 'Invalid LACP Port priority',
             211: 'Invalid LACP Time out options',
             212: 'Invalid LACP Command options',
             213: 'Invalid LLDP TLV Option',
             214: 'Invalid LLDP Option',
             215: 'Invalid Load interval delay',
             216: 'Invalid Load interval Counter Number',
             217: 'Invalid Load Interval option',
             218: 'Invalid Mac Access Group Name',
             219: 'Invalid Mac Address',
             220: 'Invalid Microburst threshold value',
             221: 'Invalid MTU Value',
             222: 'Invalid Service instance value',
             223: 'Invalid service policy name',
             224: 'Invalid service policy options',
             225: 'Invalid Interface speed value',
             226: 'Invalid Storm control level value',
             227: 'Invalid Storm control option',
             228: 'Invalid Portchannel dot1q tag',
             229: 'Invalid VRRP Id Value',
             230: 'Invalid VRRP Options',
             231: 'Invalid portchannel source interface option',
             232: 'Invalid portchannel load balance options',
             233: 'Invalid Portchannel configuration attribute',
             234: 'Invalid BFD Interval Value',
             235: 'Invalid BFD minrx Value',
             236: 'Invalid BFD multiplier Value',
             237: 'Invalid Key Chain Value',
             238: 'Invalid key name option',
             239: 'Invalid key id value',
             240: 'Invalid Key Option',
             241: 'Invalid authentication option',
             242: 'Invalid destination Ip',
             243: 'Invalid source Ip',
             244: 'Invalid IP Option',
             245: 'Invalid Access group option',
             246: 'Invalid Access group name',
             247: 'Invalid ARP MacAddress Value',
             248: 'Invalid ARP timeout value',
             249: 'Invalid ARP Option',
             250: 'Invalid dhcp request option',
             251: 'Invalid dhcp Client option',
             252: 'Invalid relay Ip Address',
             253: 'Invalid dhcp Option',
             254: 'Invalid OSPF Option',
             255: 'Invalid OSPF Id IP Address Value',
             256: 'Invalid Ip Router Option',
             257: 'Invalid Spanning tree bpdufilter Options',
             258: 'Invalid Spanning tree bpduguard Options',
             259: 'Invalid Spanning tree cost Options',
             260: 'Invalid Spanning tree guard Options',
             261: 'Invalid Spanning tree link-type Options',
             262: 'Invalid Spanning tree link-type Options',
             263: 'Invalid Spanning tree options',
             264: 'Port-priority in increments of 32 is required',
             265: 'Invalid Spanning tree vlan options',
             266: 'Invalid IPv6 option',
             267: 'Invalid IPV6 neighbor IP Address',
             268: 'Invalid IPV6 neighbor mac address',
             269: 'Invalid IPV6 dhcp option',
             270: 'Invalid IPV6 relay address option',
             271: 'Invalid IPV6 Ethernet option',
             272: 'Invalid IPV6 Vlan option',
             273: 'Invalid IPV6 Link Local option',
             274: 'Invalid IPV6 dhcp option',
             275: 'Invalid IPV6 Address',
             276: 'Invalid IPV6 Address option',
             277: 'Invalid BFD neighbor options',
             278: 'Invalid Secondary option',
             289: 'Invalid PortChannel IPV4 address',
             290: 'Invalid Max Path Options',
             291: 'Invalid Distance Local Route value',
             292: 'Invalid Distance Internal AS value',
             293: 'Invalid Distance External AS value',
             294: 'Invalid BGP Reachability Half Life',
             295: 'Invalid BGP Dampening parameter',
             296: 'Invalid BGP Aggregate Prefix value',
             297: 'Invalid BGP Aggregate Prefix Option',
             298: 'Invalid BGP Address Family Route Map Name',
             299: 'Invalid BGP Net IP Mask Value',
             300: 'Invalid BGP Net IP Prefix Value',
             301: 'Invalid BGP Neighbor configuration option',
             302: 'Invalid BGP Neighbor Weight Value',
             303: 'Invalid Neigbor update source option',
             304: 'Invalid Ethernet slot/chassis number',
             305: 'Invalid Loopback Interface number',
             306: 'Invalid vlan id',
             307: 'Invalid Number of hops',
             308: 'Invalid Neighbor Keepalive interval',
             309: 'Invalid Neighbor timer hold time',
             310: 'Invalid neighbor password ',
             311: 'Invalid Max peer limit',
             312: 'Invalid Local AS Number',
             313: 'Invalid maximum hop count',
             314: 'Invalid neighbor description',
             315: 'Invalid Neighbor connect timer value',
             316: 'Invalid Neighbor address family option',
             317: 'Invalid neighbor address family option',
             318: 'Invalid route-map name',
             319: 'Invalid route-map',
             320: 'Invalid Name of a prefix list',
             321: 'Invalid Filter incoming option',
             322: 'Invalid AS path access-list name',
             323: 'Invalid Filter route option',
             324: 'Invalid route-map name',
             325: 'Invalid Number of occurrences of AS number',
             326: 'Invalid Prefix Limit'}


def getErrorString(errorCode):
    retVal = errorDict[int(errorCode)]
    return retVal
# EOM



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017 Google
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# ----------------------------------------------------------------------------
#
#     ***     AUTO GENERATED CODE    ***    AUTO GENERATED CODE     ***
#
# ----------------------------------------------------------------------------
#
#     This file is automatically generated by Magic Modules and manual
#     changes will be clobbered when the file is regenerated.
#
#     Please read more about how to change this file at
#     https://www.github.com/GoogleCloudPlatform/magic-modules
#
# ----------------------------------------------------------------------------

from __future__ import absolute_import, division, print_function
__metaclass__ = type

################################################################################
# Documentation
################################################################################

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ["preview"],
                    'supported_by': 'community'}

DOCUMENTATION = '''
---
module: gcp_container_node_pool
description:
    - NodePool contains the name and configuration for a cluster's node pool.
    - Node pools are a set of nodes (i.e. VM's), with a common configuration and specification,
      under the control of the cluster master. They may have a set of Kubernetes labels
      applied to them, which may be used to reference them during pod scheduling. They
      may also be resized up or down, to accommodate the workload.
short_description: Creates a GCP NodePool
version_added: 2.6
author: Google Inc. (@googlecloudplatform)
requirements:
    - python >= 2.6
    - requests >= 2.18.4
    - google-auth >= 1.3.0
options:
    state:
        description:
            - Whether the given object should exist in GCP
        choices: ['present', 'absent']
        default: 'present'
    name:
        description:
            - The name of the node pool.
        required: false
    config:
        description:
            - The node configuration of the pool.
        required: false
        suboptions:
            machine_type:
                description:
                    - The name of a Google Compute Engine machine type (e.g.
                    - n1-standard-1).  If unspecified, the default machine type is n1-standard-1.
                required: false
            disk_size_gb:
                description:
                    - Size of the disk attached to each node, specified in GB. The smallest allowed disk
                      size is 10GB. If unspecified, the default disk size is 100GB.
                required: false
            oauth_scopes:
                description:
                    - The set of Google API scopes to be made available on all of the node VMs under the
                      "default" service account.
                    - 'The following scopes are recommended, but not required, and by default are not
                      included:  U(https://www.googleapis.com/auth/compute) is required for mounting persistent
                      storage on your nodes.'
                    - U(https://www.googleapis.com/auth/devstorage.read_only) is required for communicating
                      with gcr.io (the Google Container Registry).
                    - If unspecified, no scopes are added, unless Cloud Logging or Cloud Monitoring are
                      enabled, in which case their required scopes will be added.
                required: false
            service_account:
                description:
                    - The Google Cloud Platform Service Account to be used by the node VMs.  If no Service
                      Account is specified, the "default" service account is used.
                required: false
            metadata:
                description:
                    - The metadata key/value pairs assigned to instances in the cluster.
                    - 'Keys must conform to the regexp [a-zA-Z0-9-_]+ and be less than 128 bytes in length.
                      These are reflected as part of a URL in the metadata server. Additionally, to avoid
                      ambiguity, keys must not conflict with any other metadata keys for the project or
                      be one of the four reserved keys: "instance-template", "kube-env", "startup-script",
                      and "user-data"  Values are free-form strings, and only have meaning as interpreted
                      by the image running in the instance. The only restriction placed on them is that
                      each value''s size must be less than or equal to 32 KB.'
                    - The total size of all keys and values must be less than 512 KB.
                    - 'An object containing a list of "key": value pairs.'
                    - 'Example: { "name": "wrench", "mass": "1.3kg", "count": "3" }.'
                required: false
            image_type:
                description:
                    - The image type to use for this node.  Note that for a given image type, the latest
                      version of it will be used.
                required: false
            labels:
                description:
                    - 'The map of Kubernetes labels (key/value pairs) to be
                      applied to each node. These will added in addition to any
                      default label(s) that Kubernetes may apply to the node. In
                      case of conflict in label keys, the applied set may differ
                      depending on the Kubernetes version -- it''s best to assume
                      the behavior is undefined and conflicts should be avoided. For
                      more information, including usage and the valid values, see:
                      U(http://kubernetes.io/v1.1/docs/user-guide/labels.html) An
                      object containing a list of "key": value pairs.'
                    - 'Example: { "name": "wrench", "mass": "1.3kg", "count": "3" }.'
                required: false
            local_ssd_count:
                description:
                    - The number of local SSD disks to be attached to the node.
                    - 'The limit for this value is dependant upon the maximum number of disks available
                      on a machine per zone. See:  U(https://cloud.google.com/compute/docs/disks/local-ssd#local_ssd_limits)  for
                      more information.'
                required: false
            tags:
                description:
                    - The list of instance tags applied to all nodes. Tags are used to identify valid
                      sources or targets for network firewalls and are specified by the client during
                      cluster or node pool creation. Each tag within the list must comply with RFC1035.
                required: false
            preemptible:
                description:
                    - 'Whether the nodes are created as preemptible VM instances. See: U(https://cloud.google.com/compute/docs/instances/preemptible)
                      for more inforamtion about preemptible VM instances.'
                required: false
                type: bool
    initial_node_count:
        description:
            - The initial node count for the pool. You must ensure that your Compute Engine resource
              quota is sufficient for this number of instances. You must also have available firewall
              and routes quota.
        required: true
    autoscaling:
        description:
            - Autoscaler configuration for this NodePool. Autoscaler is enabled only if a valid
              configuration is present.
        required: false
        suboptions:
            enabled:
                description:
                    - Is autoscaling enabled for this node pool.
                required: false
                type: bool
            min_node_count:
                description:
                    - Minimum number of nodes in the NodePool. Must be >= 1 and <= maxNodeCount.
                required: false
            max_node_count:
                description:
                    - Maximum number of nodes in the NodePool. Must be >= minNodeCount.
                    - There has to enough quota to scale up the cluster.
                required: false
    management:
        description:
            - Management configuration for this NodePool.
        required: false
        suboptions:
            auto_upgrade:
                description:
                    - A flag that specifies whether node auto-upgrade is enabled for the node pool. If
                      enabled, node auto-upgrade helps keep the nodes in your node pool up to date with
                      the latest release version of Kubernetes.
                required: false
                type: bool
            auto_repair:
                description:
                    - A flag that specifies whether the node auto-repair is enabled for the node pool.
                      If enabled, the nodes in this node pool will be monitored and, if they fail health
                      checks too many times, an automatic repair action will be triggered.
                required: false
                type: bool
            upgrade_options:
                description:
                    - Specifies the Auto Upgrade knobs for the node pool.
                required: false
                suboptions:
                    auto_upgrade_start_time:
                        description:
                            - This field is set when upgrades are about to commence with the approximate start
                              time for the upgrades, in RFC3339 text format.
                        required: false
                    description:
                        description:
                            - This field is set when upgrades are about to commence with the description of the
                              upgrade.
                        required: false
    cluster:
        description:
            - A reference to Cluster resource.
        required: true
    zone:
        description:
            - The zone where the node pool is deployed.
        required: true
extends_documentation_fragment: gcp
'''

EXAMPLES = '''
- name: create a cluster
  gcp_container_cluster:
      name: 'cluster-nodepool'
      initial_node_count: 4
      zone: 'us-central1-a'
      project: "{{ gcp_project }}"
      auth_kind: "{{ gcp_cred_kind }}"
      service_account_file: "{{ gcp_cred_file }}"
      scopes:
        - https://www.googleapis.com/auth/cloud-platform
      state: present
  register: cluster
- name: create a node pool
  gcp_container_node_pool:
      name: testObject
      initial_node_count: 4
      cluster: "{{ cluster }}"
      zone: 'us-central1-a'
      project: testProject
      auth_kind: service_account
      service_account_file: /tmp/auth.pem
      scopes:
        - https://www.googleapis.com/auth/cloud-platform
      state: present
'''

RETURN = '''
    name:
        description:
            - The name of the node pool.
        returned: success
        type: str
    config:
        description:
            - The node configuration of the pool.
        returned: success
        type: complex
        contains:
            machine_type:
                description:
                    - The name of a Google Compute Engine machine type (e.g.
                    - n1-standard-1).  If unspecified, the default machine type is n1-standard-1.
                returned: success
                type: str
            disk_size_gb:
                description:
                    - Size of the disk attached to each node, specified in GB. The smallest allowed disk
                      size is 10GB. If unspecified, the default disk size is 100GB.
                returned: success
                type: int
            oauth_scopes:
                description:
                    - The set of Google API scopes to be made available on all of the node VMs under the
                      "default" service account.
                    - 'The following scopes are recommended, but not required, and by default are not
                      included:  U(https://www.googleapis.com/auth/compute) is required for mounting persistent
                      storage on your nodes.'
                    - U(https://www.googleapis.com/auth/devstorage.read_only) is required for communicating
                      with gcr.io (the Google Container Registry).
                    - If unspecified, no scopes are added, unless Cloud Logging or Cloud Monitoring are
                      enabled, in which case their required scopes will be added.
                returned: success
                type: list
            service_account:
                description:
                    - The Google Cloud Platform Service Account to be used by the node VMs.  If no Service
                      Account is specified, the "default" service account is used.
                returned: success
                type: str
            metadata:
                description:
                    - The metadata key/value pairs assigned to instances in the cluster.
                    - 'Keys must conform to the regexp [a-zA-Z0-9-_]+ and be less than 128 bytes in length.
                      These are reflected as part of a URL in the metadata server. Additionally, to avoid
                      ambiguity, keys must not conflict with any other metadata keys for the project or
                      be one of the four reserved keys: "instance-template", "kube-env", "startup-script",
                      and "user-data"  Values are free-form strings, and only have meaning as interpreted
                      by the image running in the instance. The only restriction placed on them is that
                      each value''s size must be less than or equal to 32 KB.'
                    - The total size of all keys and values must be less than 512 KB.
                    - 'An object containing a list of "key": value pairs.'
                    - 'Example: { "name": "wrench", "mass": "1.3kg", "count": "3" }.'
                returned: success
                type: dict
            image_type:
                description:
                    - The image type to use for this node.  Note that for a given image type, the latest
                      version of it will be used.
                returned: success
                type: str
            labels:
                description:
                    - 'The map of Kubernetes labels (key/value pairs) to be
                      applied to each node. These will added in addition to any
                      default label(s) that Kubernetes may apply to the node. In
                      case of conflict in label keys, the applied set may differ
                      depending on the Kubernetes version -- it''s best to assume
                      the behavior is undefined and conflicts should be avoided. For
                      more information, including usage and the valid values, see:
                      U(http://kubernetes.io/v1.1/docs/user-guide/labels.html) An
                      object containing a list of "key": value pairs.'
                    - 'Example: { "name": "wrench", "mass": "1.3kg", "count": "3" }.'
                returned: success
                type: dict
            local_ssd_count:
                description:
                    - The number of local SSD disks to be attached to the node.
                    - 'The limit for this value is dependant upon the maximum number of disks available
                      on a machine per zone. See:  U(https://cloud.google.com/compute/docs/disks/local-ssd#local_ssd_limits)  for
                      more information.'
                returned: success
                type: int
            tags:
                description:
                    - The list of instance tags applied to all nodes. Tags are used to identify valid
                      sources or targets for network firewalls and are specified by the client during
                      cluster or node pool creation. Each tag within the list must comply with RFC1035.
                returned: success
                type: list
            preemptible:
                description:
                    - 'Whether the nodes are created as preemptible VM instances. See: U(https://cloud.google.com/compute/docs/instances/preemptible)
                      for more inforamtion about preemptible VM instances.'
                returned: success
                type: bool
    initial_node_count:
        description:
            - The initial node count for the pool. You must ensure that your Compute Engine resource
              quota is sufficient for this number of instances. You must also have available firewall
              and routes quota.
        returned: success
        type: int
    version:
        description:
            - The version of the Kubernetes of this node.
        returned: success
        type: str
    autoscaling:
        description:
            - Autoscaler configuration for this NodePool. Autoscaler is enabled only if a valid
              configuration is present.
        returned: success
        type: complex
        contains:
            enabled:
                description:
                    - Is autoscaling enabled for this node pool.
                returned: success
                type: bool
            min_node_count:
                description:
                    - Minimum number of nodes in the NodePool. Must be >= 1 and <= maxNodeCount.
                returned: success
                type: int
            max_node_count:
                description:
                    - Maximum number of nodes in the NodePool. Must be >= minNodeCount.
                    - There has to enough quota to scale up the cluster.
                returned: success
                type: int
    management:
        description:
            - Management configuration for this NodePool.
        returned: success
        type: complex
        contains:
            auto_upgrade:
                description:
                    - A flag that specifies whether node auto-upgrade is enabled for the node pool. If
                      enabled, node auto-upgrade helps keep the nodes in your node pool up to date with
                      the latest release version of Kubernetes.
                returned: success
                type: bool
            auto_repair:
                description:
                    - A flag that specifies whether the node auto-repair is enabled for the node pool.
                      If enabled, the nodes in this node pool will be monitored and, if they fail health
                      checks too many times, an automatic repair action will be triggered.
                returned: success
                type: bool
            upgrade_options:
                description:
                    - Specifies the Auto Upgrade knobs for the node pool.
                returned: success
                type: complex
                contains:
                    auto_upgrade_start_time:
                        description:
                            - This field is set when upgrades are about to commence with the approximate start
                              time for the upgrades, in RFC3339 text format.
                        returned: success
                        type: str
                    description:
                        description:
                            - This field is set when upgrades are about to commence with the description of the
                              upgrade.
                        returned: success
                        type: str
    cluster:
        description:
            - A reference to Cluster resource.
        returned: success
        type: dict
    zone:
        description:
            - The zone where the node pool is deployed.
        returned: success
        type: str
'''

################################################################################
# Imports
################################################################################

from ansible.module_utils.gcp_utils import navigate_hash, GcpSession, GcpModule, GcpRequest, remove_nones_from_dict, replace_resource_dict
import json
import time

################################################################################
# Main
################################################################################


def main():
    """Main function"""

    module = GcpModule(
        argument_spec=dict(
            state=dict(default='present', choices=['present', 'absent'], type='str'),
            name=dict(type='str'),
            config=dict(type='dict', options=dict(
                machine_type=dict(type='str'),
                disk_size_gb=dict(type='int'),
                oauth_scopes=dict(type='list', elements='str'),
                service_account=dict(type='str'),
                metadata=dict(type='dict'),
                image_type=dict(type='str'),
                labels=dict(type='dict'),
                local_ssd_count=dict(type='int'),
                tags=dict(type='list', elements='str'),
                preemptible=dict(type='bool')
            )),
            initial_node_count=dict(required=True, type='int'),
            autoscaling=dict(type='dict', options=dict(
                enabled=dict(type='bool'),
                min_node_count=dict(type='int'),
                max_node_count=dict(type='int')
            )),
            management=dict(type='dict', options=dict(
                auto_upgrade=dict(type='bool'),
                auto_repair=dict(type='bool'),
                upgrade_options=dict(type='dict', options=dict(
                    auto_upgrade_start_time=dict(type='str'),
                    description=dict(type='str')
                ))
            )),
            cluster=dict(required=True, type='dict'),
            zone=dict(required=True, type='str')
        )
    )

    state = module.params['state']

    fetch = fetch_resource(module, self_link(module))
    changed = False

    if fetch:
        if state == 'present':
            if is_different(module, fetch):
                fetch = update(module, self_link(module))
                changed = True
        else:
            delete(module, self_link(module))
            fetch = {}
            changed = True
    else:
        if state == 'present':
            fetch = create(module, collection(module))
            changed = True
        else:
            fetch = {}

    fetch.update({'changed': changed})

    module.exit_json(**fetch)


def create(module, link):
    auth = GcpSession(module, 'container')
    return wait_for_operation(module, auth.post(link, resource_to_request(module)))


def update(module, link):
    auth = GcpSession(module, 'container')
    return wait_for_operation(module, auth.put(link, resource_to_request(module)))


def delete(module, link):
    auth = GcpSession(module, 'container')
    return wait_for_operation(module, auth.delete(link))


def resource_to_request(module):
    request = {
        u'name': module.params.get('name'),
        u'config': NodePoolConfig(module.params.get('config', {}), module).to_request(),
        u'initialNodeCount': module.params.get('initial_node_count'),
        u'autoscaling': NodePoolAutosca(module.params.get('autoscaling', {}), module).to_request(),
        u'management': NodePoolManagem(module.params.get('management', {}), module).to_request()
    }
    request = encode_request(request, module)
    return_vals = {}
    for k, v in request.items():
        if v:
            return_vals[k] = v

    return return_vals


def fetch_resource(module, link):
    auth = GcpSession(module, 'container')
    return return_if_object(module, auth.get(link))


def self_link(module):
    res = {
        'project': module.params['project'],
        'zone': module.params['zone'],
        'cluster': replace_resource_dict(module.params['cluster'], 'name'),
        'name': module.params['name']
    }
    return "https://container.googleapis.com/v1/projects/{project}/zones/{zone}/clusters/{cluster}/nodePools/{name}".format(**res)


def collection(module):
    res = {
        'project': module.params['project'],
        'zone': module.params['zone'],
        'cluster': replace_resource_dict(module.params['cluster'], 'name')
    }
    return "https://container.googleapis.com/v1/projects/{project}/zones/{zone}/clusters/{cluster}/nodePools".format(**res)


def return_if_object(module, response):
    # If not found, return nothing.
    if response.status_code == 404:
        return None

    # If no content, return nothing.
    if response.status_code == 204:
        return None

    try:
        module.raise_for_status(response)
        result = response.json()
    except getattr(json.decoder, 'JSONDecodeError', ValueError) as inst:
        module.fail_json(msg="Invalid JSON response with error: %s" % inst)

    if navigate_hash(result, ['error', 'errors']):
        module.fail_json(msg=navigate_hash(result, ['error', 'errors']))

    return result


def is_different(module, response):
    request = resource_to_request(module)
    response = response_to_hash(module, response)

    # Remove all output-only from response.
    response_vals = {}
    for k, v in response.items():
        if k in request:
            response_vals[k] = v

    request_vals = {}
    for k, v in request.items():
        if k in response:
            request_vals[k] = v

    return GcpRequest(request_vals) != GcpRequest(response_vals)


# Remove unnecessary properties from the response.
# This is for doing comparisons with Ansible's current parameters.
def response_to_hash(module, response):
    return {
        u'name': response.get(u'name'),
        u'config': NodePoolConfig(response.get(u'config', {}), module).from_response(),
        u'initialNodeCount': module.params.get('initial_node_count'),
        u'version': response.get(u'version'),
        u'autoscaling': NodePoolAutosca(response.get(u'autoscaling', {}), module).from_response(),
        u'management': NodePoolManagem(response.get(u'management', {}), module).from_response()
    }


def async_op_url(module, extra_data=None):
    if extra_data is None:
        extra_data = {}
    url = "https://container.googleapis.com/v1/projects/{project}/zones/{zone}/operations/{op_id}"
    combined = extra_data.copy()
    combined.update(module.params)
    return url.format(**combined)


def wait_for_operation(module, response):
    op_result = return_if_object(module, response)
    if op_result is None:
        return None
    status = navigate_hash(op_result, ['status'])
    wait_done = wait_for_completion(status, op_result, module)
    return fetch_resource(module, navigate_hash(wait_done, ['targetLink']))


def wait_for_completion(status, op_result, module):
    op_id = navigate_hash(op_result, ['name'])
    op_uri = async_op_url(module, {'op_id': op_id})
    while status != 'DONE':
        raise_if_errors(op_result, ['error', 'errors'], 'message')
        time.sleep(1.0)
        if status not in ['PENDING', 'RUNNING', 'DONE', 'ABORTING']:
            module.fail_json(msg="Invalid result %s" % status)
        op_result = fetch_resource(module, op_uri)
        status = navigate_hash(op_result, ['status'])
    return op_result


def raise_if_errors(response, err_path, module):
    errors = navigate_hash(response, err_path)
    if errors is not None:
        module.fail_json(msg=errors)


# Google Container Engine API has its own layout for the create method,
# defined like this:
#
# {
#   'nodePool': {
#     ... node pool data
#   }
# }
#
# Format the request to match the expected input by the API
def encode_request(resource_request, module):
    return {
        'nodePool': resource_request
    }


class NodePoolConfig(object):
    def __init__(self, request, module):
        self.module = module
        if request:
            self.request = request
        else:
            self.request = {}

    def to_request(self):
        return remove_nones_from_dict({
            u'machineType': self.request.get('machine_type'),
            u'diskSizeGb': self.request.get('disk_size_gb'),
            u'oauthScopes': self.request.get('oauth_scopes'),
            u'serviceAccount': self.request.get('service_account'),
            u'metadata': self.request.get('metadata'),
            u'imageType': self.request.get('image_type'),
            u'labels': self.request.get('labels'),
            u'localSsdCount': self.request.get('local_ssd_count'),
            u'tags': self.request.get('tags'),
            u'preemptible': self.request.get('preemptible')
        })

    def from_response(self):
        return remove_nones_from_dict({
            u'machineType': self.request.get(u'machineType'),
            u'diskSizeGb': self.request.get(u'diskSizeGb'),
            u'oauthScopes': self.request.get(u'oauthScopes'),
            u'serviceAccount': self.request.get(u'serviceAccount'),
            u'metadata': self.request.get(u'metadata'),
            u'imageType': self.request.get(u'imageType'),
            u'labels': self.request.get(u'labels'),
            u'localSsdCount': self.request.get(u'localSsdCount'),
            u'tags': self.request.get(u'tags'),
            u'preemptible': self.request.get(u'preemptible')
        })


class NodePoolAutosca(object):
    def __init__(self, request, module):
        self.module = module
        if request:
            self.request = request
        else:
            self.request = {}

    def to_request(self):
        return remove_nones_from_dict({
            u'enabled': self.request.get('enabled'),
            u'minNodeCount': self.request.get('min_node_count'),
            u'maxNodeCount': self.request.get('max_node_count')
        })

    def from_response(self):
        return remove_nones_from_dict({
            u'enabled': self.request.get(u'enabled'),
            u'minNodeCount': self.request.get(u'minNodeCount'),
            u'maxNodeCount': self.request.get(u'maxNodeCount')
        })


class NodePoolManagem(object):
    def __init__(self, request, module):
        self.module = module
        if request:
            self.request = request
        else:
            self.request = {}

    def to_request(self):
        return remove_nones_from_dict({
            u'autoUpgrade': self.request.get('auto_upgrade'),
            u'autoRepair': self.request.get('auto_repair'),
            u'upgradeOptions': NodePoolUpgraOptio(self.request.get('upgrade_options', {}), self.module).to_request()
        })

    def from_response(self):
        return remove_nones_from_dict({
            u'autoUpgrade': self.request.get(u'autoUpgrade'),
            u'autoRepair': self.request.get(u'autoRepair'),
            u'upgradeOptions': NodePoolUpgraOptio(self.request.get(u'upgradeOptions', {}), self.module).from_response()
        })


class NodePoolUpgraOptio(object):
    def __init__(self, request, module):
        self.module = module
        if request:
            self.request = request
        else:
            self.request = {}

    def to_request(self):
        return remove_nones_from_dict({
            u'autoUpgradeStartTime': self.request.get('auto_upgrade_start_time'),
            u'description': self.request.get('description')
        })

    def from_response(self):
        return remove_nones_from_dict({
            u'autoUpgradeStartTime': self.request.get(u'autoUpgradeStartTime'),
            u'description': self.request.get(u'description')
        })

if __name__ == '__main__':
    main()



#!/usr/bin/env python

import optparse
import os
import pprint
import sys

from jinja2 import Environment, FileSystemLoader

from ansible.module_utils._text import to_bytes


def generate_parser():
    p = optparse.OptionParser(
        version='%prog 1.0',
        usage='usage: %prog [options]',
        description='Generate cli documentation from cli docstrings',
    )

    p.add_option("-t", "--template-file", action="store", dest="template_file", default="../templates/man.j2", help="path to jinja2 template")
    p.add_option("-o", "--output-dir", action="store", dest="output_dir", default='/tmp/', help="Output directory for rst files")
    p.add_option("-f", "--output-format", action="store", dest="output_format", default='man', help="Output format for docs (the default 'man' or 'rst')")
    return p


# from https://www.python.org/dev/peps/pep-0257/
def trim_docstring(docstring):
    if not docstring:
        return ''
    # Convert tabs to spaces (following the normal Python rules)
    # and split into a list of lines:
    lines = docstring.expandtabs().splitlines()
    # Determine minimum indentation (first line doesn't count):
    indent = sys.maxsize
    for line in lines[1:]:
        stripped = line.lstrip()
        if stripped:
            indent = min(indent, len(line) - len(stripped))
    # Remove indentation (first line is special):
    trimmed = [lines[0].strip()]
    if indent < sys.maxsize:
        for line in lines[1:]:
            trimmed.append(line[indent:].rstrip())
    # Strip off trailing and leading blank lines:
    while trimmed and not trimmed[-1]:
        trimmed.pop()
    while trimmed and not trimmed[0]:
        trimmed.pop(0)
    # Return a single string:
    return '\n'.join(trimmed)


def get_options(optlist):
    ''' get actual options '''

    opts = []
    for opt in optlist:
        res = {
            'desc': opt.help,
            'options': opt._short_opts + opt._long_opts
        }
        if opt.action == 'store':
            res['arg'] = opt.dest.upper()
        opts.append(res)

    return opts


def get_option_groups(option_parser):
    groups = []
    for option_group in option_parser.option_groups:
        group_info = {}
        group_info['desc'] = option_group.get_description()
        group_info['options'] = option_group.option_list
        group_info['group_obj'] = option_group
        groups.append(group_info)
    return groups


def opt_doc_list(cli):
    ''' iterate over options lists '''

    results = []
    for option_group in cli.parser.option_groups:
        results.extend(get_options(option_group.option_list))

    results.extend(get_options(cli.parser.option_list))

    return results


# def opts_docs(cli, name):
def opts_docs(cli_class_name, cli_module_name):
    ''' generate doc structure from options '''

    cli_name = 'ansible-%s' % cli_module_name
    if cli_module_name == 'adhoc':
        cli_name = 'ansible'

    # WIth no action/subcommand
    # shared opts set
    # instantiate each cli and ask its options
    cli_klass = getattr(__import__("ansible.cli.%s" % cli_module_name,
                                   fromlist=[cli_class_name]), cli_class_name)
    cli = cli_klass([])

    # parse the common options
    try:
        cli.parse()
    except:
        pass

    # base/common cli info
    docs = {
        'cli': cli_module_name,
        'cli_name': cli_name,
        'usage': cli.parser.usage,
        'short_desc': cli.parser.description,
        'long_desc': trim_docstring(cli.__doc__),
        'actions': {},
    }
    option_info = {'option_names': [],
                   'options': [],
                   'groups': []}

    for extras in ('ARGUMENTS'):
        if hasattr(cli, extras):
            docs[extras.lower()] = getattr(cli, extras)

    common_opts = opt_doc_list(cli)
    groups_info = get_option_groups(cli.parser)
    shared_opt_names = []
    for opt in common_opts:
        shared_opt_names.extend(opt.get('options', []))

    option_info['options'] = common_opts
    option_info['option_names'] = shared_opt_names

    option_info['groups'].extend(groups_info)

    docs.update(option_info)

    # now for each action/subcommand
    # force populate parser with per action options

    # use class attrs not the attrs on a instance (not that it matters here...)
    for action in getattr(cli_klass, 'VALID_ACTIONS', ()):
        # instantiate each cli and ask its options
        action_cli_klass = getattr(__import__("ansible.cli.%s" % cli_module_name,
                                              fromlist=[cli_class_name]), cli_class_name)
        # init with args with action added?
        cli = action_cli_klass([])
        cli.args.append(action)

        try:
            cli.parse()
        except:
            pass

        # FIXME/TODO: needed?
        # avoid dupe errors
        cli.parser.set_conflict_handler('resolve')

        cli.set_action()

        action_info = {'option_names': [],
                       'options': []}
        # docs['actions'][action] = {}
        # docs['actions'][action]['name'] = action
        action_info['name'] = action
        action_info['desc'] = trim_docstring(getattr(cli, 'execute_%s' % action).__doc__)

        # docs['actions'][action]['desc'] = getattr(cli, 'execute_%s' % action).__doc__.strip()
        action_doc_list = opt_doc_list(cli)

        uncommon_options = []
        for action_doc in action_doc_list:
            # uncommon_options = []

            option_aliases = action_doc.get('options', [])
            for option_alias in option_aliases:

                if option_alias in shared_opt_names:
                    continue

                # TODO: use set
                if option_alias not in action_info['option_names']:
                    action_info['option_names'].append(option_alias)

                if action_doc in action_info['options']:
                    continue

                uncommon_options.append(action_doc)

            action_info['options'] = uncommon_options

        docs['actions'][action] = action_info

    docs['options'] = opt_doc_list(cli)
    return docs


if __name__ == '__main__':

    parser = generate_parser()

    options, args = parser.parse_args()

    template_file = options.template_file
    template_path = os.path.expanduser(template_file)
    template_dir = os.path.abspath(os.path.dirname(template_path))
    template_basename = os.path.basename(template_file)

    output_dir = os.path.abspath(options.output_dir)
    output_format = options.output_format

    cli_modules = args

    # various cli parsing things checks sys.argv if the 'args' that are passed in are []
    # so just remove any args so the cli modules dont try to parse them resulting in warnings
    sys.argv = [sys.argv[0]]
    # need to be in right dir
    os.chdir(os.path.dirname(__file__))

    allvars = {}
    output = {}
    cli_list = []
    cli_bin_name_list = []

    # for binary in os.listdir('../../lib/ansible/cli'):
    for cli_module_name in cli_modules:
        binary = os.path.basename(os.path.expanduser(cli_module_name))

        if not binary.endswith('.py'):
            continue
        elif binary == '__init__.py':
            continue

        cli_name = os.path.splitext(binary)[0]

        if cli_name == 'adhoc':
            cli_class_name = 'AdHocCLI'
            # myclass = 'AdHocCLI'
            output[cli_name] = 'ansible.1.rst.in'
            cli_bin_name = 'ansible'
        else:
            # myclass = "%sCLI" % libname.capitalize()
            cli_class_name = "%sCLI" % cli_name.capitalize()
            output[cli_name] = 'ansible-%s.1.rst.in' % cli_name
            cli_bin_name = 'ansible-%s' % cli_name

        # FIXME:
        allvars[cli_name] = opts_docs(cli_class_name, cli_name)
        cli_bin_name_list.append(cli_bin_name)

    cli_list = allvars.keys()

    doc_name_formats = {'man': '%s.1.rst.in',
                        'rst': '%s.rst'}

    for cli_name in cli_list:

        # template it!
        env = Environment(loader=FileSystemLoader(template_dir))
        template = env.get_template(template_basename)

        # add rest to vars
        tvars = allvars[cli_name]
        tvars['cli_list'] = cli_list
        tvars['cli_bin_name_list'] = cli_bin_name_list
        tvars['cli'] = cli_name
        if '-i' in tvars['options']:
            print('uses inventory')

        manpage = template.render(tvars)
        filename = os.path.join(output_dir, doc_name_formats[output_format] % tvars['cli_name'])

        with open(filename, 'wb') as f:
            f.write(to_bytes(manpage))
            print("Wrote doc to %s" % filename)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# (c) 2018, Ansible by Red Hat, inc
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'network'}


DOCUMENTATION = """
---
module: net_put
version_added: "2.6"
author: "Deepak Agrawal (@dagrawal)"
short_description: Copy a file from Ansible Controller to a network device
description:
  - This module provides functionality to copy file from Ansible controller to
    network devices.
options:
  src:
    description:
      - Specifies the source file. The path to the source file can either be
        the full path on the Ansible control host or a relative path from the
        playbook or role root directory.
    required: true
  protocol:
    description:
      - Protocol used to transfer file.
    default: scp
    choices: ['scp', 'sftp']
  dest:
    description:
      - Specifies the destination file. The path to destination file can
        either be the full path or relative path as supported by network_os.
    default:
      - Filename from src and at default directory of user shell on
        network_os.
    required: no

requirements:
    - "scp"

notes:
   - Some devices need specific configurations to be enabled before scp can work
     These configuration should be pre-configured before using this module
     e.g ios - C(ip scp server enable).
   - User privilege to do scp on network device should be pre-configured
     e.g. ios - need user privilege 15 by default for allowing scp.
   - Default destination of source file.
"""

EXAMPLES = """
- name: copy file from ansible controller to a network device
  net_put:
    src: running_cfg_ios1.txt

- name: copy file at root dir of flash in slot 3 of sw1(ios)
  net_put:
    src: running_cfg_sw1.txt
    protocol: sftp
    dest : flash3:/running_cfg_sw1.txt
"""

RETURN = """
"""



# (c) 2017, Brian Coca <bcoca@ansible.com>
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

import optparse
from operator import attrgetter

from ansible import constants as C
from ansible.cli import CLI
from ansible.errors import AnsibleError, AnsibleOptionsError
from ansible.inventory.host import Host
from ansible.plugins.loader import vars_loader
from ansible.parsing.dataloader import DataLoader
from ansible.utils.vars import combine_vars

try:
    from __main__ import display
except ImportError:
    from ansible.utils.display import Display
    display = Display()

INTERNAL_VARS = frozenset(['ansible_diff_mode',
                           'ansible_facts',
                           'ansible_forks',
                           'ansible_inventory_sources',
                           'ansible_limit',
                           'ansible_playbook_python',
                           'ansible_run_tags',
                           'ansible_skip_tags',
                           'ansible_verbosity',
                           'ansible_version',
                           'inventory_dir',
                           'inventory_file',
                           'inventory_hostname',
                           'inventory_hostname_short',
                           'groups',
                           'group_names',
                           'omit',
                           'playbook_dir', ])


class InventoryCLI(CLI):
    ''' used to display or dump the configured inventory as Ansible sees it '''

    ARGUMENTS = {'host': 'The name of a host to match in the inventory, relevant when using --list',
                 'group': 'The name of a group in the inventory, relevant when using --graph', }

    def __init__(self, args):

        super(InventoryCLI, self).__init__(args)
        self.vm = None
        self.loader = None
        self.inventory = None

        self._new_api = True

    def parse(self):

        self.parser = CLI.base_parser(
            usage='usage: %prog [options] [host|group]',
            epilog='Show Ansible inventory information, by default it uses the inventory script JSON format',
            inventory_opts=True,
            vault_opts=True,
            basedir_opts=True,
        )

        # remove unused default options
        self.parser.remove_option('--limit')
        self.parser.remove_option('--list-hosts')

        # Actions
        action_group = optparse.OptionGroup(self.parser, "Actions", "One of following must be used on invocation, ONLY ONE!")
        action_group.add_option("--list", action="store_true", default=False, dest='list', help='Output all hosts info, works as inventory script')
        action_group.add_option("--host", action="store", default=None, dest='host', help='Output specific host info, works as inventory script')
        action_group.add_option("--graph", action="store_true", default=False, dest='graph',
                                help='create inventory graph, if supplying pattern it must be a valid group name')
        self.parser.add_option_group(action_group)

        # graph
        self.parser.add_option("-y", "--yaml", action="store_true", default=False, dest='yaml',
                               help='Use YAML format instead of default JSON, ignored for --graph')
        self.parser.add_option("--vars", action="store_true", default=False, dest='show_vars',
                               help='Add vars to graph display, ignored unless used with --graph')

        # list
        self.parser.add_option("--export", action="store_true", default=C.INVENTORY_EXPORT, dest='export',
                               help="When doing an --list, represent in a way that is optimized for export,"
                                    "not as an accurate representation of how Ansible has processed it")
        # self.parser.add_option("--ignore-vars-plugins", action="store_true", default=False, dest='ignore_vars_plugins',
        #                       help="When doing an --list, skip vars data from vars plugins, by default, this would include group_vars/ and host_vars/")

        super(InventoryCLI, self).parse()

        display.verbosity = self.options.verbosity

        self.validate_conflicts(vault_opts=True)

        # there can be only one! and, at least, one!
        used = 0
        for opt in (self.options.list, self.options.host, self.options.graph):
            if opt:
                used += 1
        if used == 0:
            raise AnsibleOptionsError("No action selected, at least one of --host, --graph or --list needs to be specified.")
        elif used > 1:
            raise AnsibleOptionsError("Conflicting options used, only one of --host, --graph or --list can be used at the same time.")

        # set host pattern to default if not supplied
        if len(self.args) > 0:
            self.options.pattern = self.args[0]
        else:
            self.options.pattern = 'all'

    def run(self):

        results = None

        super(InventoryCLI, self).run()

        # Initialize needed objects
        if getattr(self, '_play_prereqs', False):
            self.loader, self.inventory, self.vm = self._play_prereqs(self.options)
        else:
            # fallback to pre 2.4 way of initialzing
            from ansible.vars import VariableManager
            from ansible.inventory import Inventory

            self._new_api = False
            self.loader = DataLoader()
            self.vm = VariableManager()

            # use vault if needed
            if self.options.vault_password_file:
                vault_pass = CLI.read_vault_password_file(self.options.vault_password_file, loader=self.loader)
            elif self.options.ask_vault_pass:
                vault_pass = self.ask_vault_passwords()
            else:
                vault_pass = None

            if vault_pass:
                self.loader.set_vault_password(vault_pass)
                # actually get inventory and vars

            self.inventory = Inventory(loader=self.loader, variable_manager=self.vm, host_list=self.options.inventory)
            self.vm.set_inventory(self.inventory)

        if self.options.host:
            hosts = self.inventory.get_hosts(self.options.host)
            if len(hosts) != 1:
                raise AnsibleOptionsError("You must pass a single valid host to --hosts parameter")

            myvars = self._get_host_variables(host=hosts[0])
            self._remove_internal(myvars)

            # FIXME: should we template first?
            results = self.dump(myvars)

        elif self.options.graph:
            results = self.inventory_graph()
        elif self.options.list:
            top = self._get_group('all')
            if self.options.yaml:
                results = self.yaml_inventory(top)
            else:
                results = self.json_inventory(top)
            results = self.dump(results)

        if results:
            # FIXME: pager?
            display.display(results)
            exit(0)

        exit(1)

    def dump(self, stuff):

        if self.options.yaml:
            import yaml
            from ansible.parsing.yaml.dumper import AnsibleDumper
            results = yaml.dump(stuff, Dumper=AnsibleDumper, default_flow_style=False)
        else:
            import json
            from ansible.parsing.ajson import AnsibleJSONEncoder
            results = json.dumps(stuff, cls=AnsibleJSONEncoder, sort_keys=True, indent=4)

        return results

    # FIXME: refactor to use same for VM
    def get_plugin_vars(self, path, entity):

        data = {}

        def _get_plugin_vars(plugin, path, entities):
            data = {}
            try:
                data = plugin.get_vars(self.loader, path, entity)
            except AttributeError:
                try:
                    if isinstance(entity, Host):
                        data = combine_vars(data, plugin.get_host_vars(entity.name))
                    else:
                        data = combine_vars(data, plugin.get_group_vars(entity.name))
                except AttributeError:
                    if hasattr(plugin, 'run'):
                        raise AnsibleError("Cannot use v1 type vars plugin %s from %s" % (plugin._load_name, plugin._original_path))
                    else:
                        raise AnsibleError("Invalid vars plugin %s from %s" % (plugin._load_name, plugin._original_path))
            return data

        for plugin in vars_loader.all():
            data = combine_vars(data, _get_plugin_vars(plugin, path, entity))

        return data

    def _get_group_variables(self, group):

        # get info from inventory source
        res = group.get_vars()

        # FIXME: add switch to skip vars plugins, add vars plugin info
        for inventory_dir in self.inventory._sources:
            res = combine_vars(res, self.get_plugin_vars(inventory_dir, group))

        if group.priority != 1:
            res['ansible_group_priority'] = group.priority

        return res

    def _get_host_variables(self, host):

        if self.options.export:
            hostvars = host.get_vars()

            # FIXME: add switch to skip vars plugins
            # add vars plugin info
            for inventory_dir in self.inventory._sources:
                hostvars = combine_vars(hostvars, self.get_plugin_vars(inventory_dir, host))
        else:
            if self._new_api:
                hostvars = self.vm.get_vars(host=host, include_hostvars=False)
            else:
                hostvars = self.vm.get_vars(self.loader, host=host, include_hostvars=False)

        return hostvars

    def _get_group(self, gname):
        if self._new_api:
            group = self.inventory.groups.get(gname)
        else:
            group = self.inventory.get_group(gname)
        return group

    def _remove_internal(self, dump):

        for internal in INTERNAL_VARS:
            if internal in dump:
                del dump[internal]

    def _remove_empty(self, dump):
        # remove empty keys
        for x in ('hosts', 'vars', 'children'):
            if x in dump and not dump[x]:
                del dump[x]

    def _show_vars(self, dump, depth):
        result = []
        self._remove_internal(dump)
        if self.options.show_vars:
            for (name, val) in sorted(dump.items()):
                result.append(self._graph_name('{%s = %s}' % (name, val), depth))
        return result

    def _graph_name(self, name, depth=0):
        if depth:
            name = "  |" * (depth) + "--%s" % name
        return name

    def _graph_group(self, group, depth=0):

        result = [self._graph_name('@%s:' % group.name, depth)]
        depth = depth + 1
        for kid in sorted(group.child_groups, key=attrgetter('name')):
            result.extend(self._graph_group(kid, depth))

        if group.name != 'all':
            for host in sorted(group.hosts, key=attrgetter('name')):
                result.append(self._graph_name(host.name, depth))
                result.extend(self._show_vars(host.get_vars(), depth + 1))

        result.extend(self._show_vars(self._get_group_variables(group), depth))

        return result

    def inventory_graph(self):

        start_at = self._get_group(self.options.pattern)
        if start_at:
            return '\n'.join(self._graph_group(start_at))
        else:
            raise AnsibleOptionsError("Pattern must be valid group name when using --graph")

    def json_inventory(self, top):

        def format_group(group):
            results = {}
            results[group.name] = {}
            if group.name != 'all':
                results[group.name]['hosts'] = [h.name for h in sorted(group.hosts, key=attrgetter('name'))]
            results[group.name]['children'] = []
            for subgroup in sorted(group.child_groups, key=attrgetter('name')):
                results[group.name]['children'].append(subgroup.name)
                results.update(format_group(subgroup))
            if self.options.export:
                results[group.name]['vars'] = self._get_group_variables(group)

            self._remove_empty(results[group.name])

            return results

        results = format_group(top)

        # populate meta
        results['_meta'] = {'hostvars': {}}
        hosts = self.inventory.get_hosts()
        for host in hosts:
            hvars = self._get_host_variables(host)
            if hvars:
                self._remove_internal(hvars)
                results['_meta']['hostvars'][host.name] = hvars

        return results

    def yaml_inventory(self, top):

        seen = []

        def format_group(group):
            results = {}

            # initialize group + vars
            results[group.name] = {}

            # subgroups
            results[group.name]['children'] = {}
            for subgroup in sorted(group.child_groups, key=attrgetter('name')):
                if subgroup.name != 'all':
                    results[group.name]['children'].update(format_group(subgroup))

            # hosts for group
            results[group.name]['hosts'] = {}
            if group.name != 'all':
                for h in sorted(group.hosts, key=attrgetter('name')):
                    myvars = {}
                    if h.name not in seen:  # avoid defining host vars more than once
                        seen.append(h.name)
                        myvars = self._get_host_variables(host=h)
                        self._remove_internal(myvars)
                    results[group.name]['hosts'][h.name] = myvars

            if self.options.export:

                gvars = self._get_group_variables(group)
                if gvars:
                    results[group.name]['vars'] = gvars

            self._remove_empty(results[group.name])

            return results

        return format_group(top)
